#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" 
Collection of functions to segment the sound of interests (also called regions 
of interest (ROIs) in time and frequency.
"""
#
# Authors:  Felix MICHAUD   <felixmichaudlnhrdt@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

#%%

# basic packages
import numpy as np
import matplotlib.pyplot as plt

# scikit-image (image processing) package
from skimage.transform import resize

# Scikit-Maad (ecoacoustics functions) package
import maad

# import bambird
from bambird import config as cfg

#%%
###############################################################################
def _intersection_bbox(bb1, bb2):
    """
    test if 2 bounding-boxes intersect.

    Parameters
    ----------
    bb1 : dict
        Keys: {'x1', 'x2', 'y1', 'y2'}
        The (x1, y1) position is at the top left corner,
        the (x2, y2) position is at the bottom right corner
    bb2 : dict
        Keys: {'x1', 'x2', 'y1', 'y2'}
        The (x, y) position is at the top left corner,
        the (x2, y2) position is at the bottom right corner

    Returns
    -------
    boolean

    """

    # test if values are consistant (min < max)
    assert bb1["min_x"] <= bb1["max_x"]
    assert bb1["min_y"] <= bb1["max_y"]
    assert bb2["min_x"] <= bb2["max_x"]
    assert bb2["min_y"] <= bb2["max_y"]

    # determine the coordinates of the intersection rectangle
    x_left = max(bb1["min_x"], bb2["min_x"])
    y_top = max(bb1["min_y"], bb2["min_y"])
    x_right = min(bb1["max_x"], bb2["max_x"])
    y_bottom = min(bb1["max_y"], bb2["max_y"])

    if (x_right < x_left) or (y_bottom < y_top):
        is_intersected = False
    else:
        is_intersected = True

    return is_intersected

###############################################################################
def _fusion_bbox(df_rois, idx, idx2):
    """[fusion beginning and ends of 2 events that are too close to each other. It is applied if
    the 2 beginnings or the 2 ends are too close, or if the end of the first one is too close to the beginning
    of the second event. This function test all the events that follow each other in the dataframe.]

    Args:
        df_rois : 
            dataframe with all bbox to test for merging
        idx1 : 
            position of the FIRST bbox in the dataframe
        idx2 : 
            position of the SECOND bbox in the dataframe

    Return       df_rois: 
            returns the original dataframe with merged bbox
    """
    bb1 = df_rois.loc[idx]
    bb2 = df_rois.loc[idx2]

    min_t = min(bb1.min_t, bb2.min_t)
    max_t = max(bb1.max_t, bb2.max_t)
    min_f = min(bb1.min_f, bb2.min_f)
    max_f = max(bb1.max_f, bb2.max_f)

    df_rois.loc[idx, ["min_f", "min_t", "max_f", "max_t"]] = [
        min_f,
        min_t,
        max_f,
        max_t,
    ]

    min_x = min(bb1.min_x, bb2.min_x)
    max_x = max(bb1.max_x, bb2.max_x)
    min_y = min(bb1.min_y, bb2.min_y)
    max_y = max(bb1.max_y, bb2.max_y)

    df_rois.loc[idx, ["min_y", "min_x", "max_y", "max_x"]] = [
        min_y,
        min_x,
        max_y,
        max_x,
    ]

    df_rois = df_rois.drop(index=idx2)

    return df_rois

###############################################################################
def _merge_bbox(df_rois, margins):
    """[Merge two bbox that are within the margin. As the process is iterative
    if the resulting bbox is also close to another bbox, then it is merged. 
    And so on]

    Args:
        df_rois : 
            dataframe with all bbox to test for merging
        margins : 
            array with margins before/after the bbox and upper/lower the bbox

    Returns:
        df_rois: 
            returns the original dataframe with merged bbox
    """

    # loop until all bbox are merged
    # test if the number of rois is stable. if not, then run again the merging
    # process.
    num_rois = len(df_rois) + 1
    while len(df_rois) < num_rois:
        num_rois = len(df_rois)
        for idx, row in df_rois.iterrows():
            # test if the index idx is still in the dataframe
            if idx in df_rois.index:
                # remove the current ROIs (to avoid auto-merge)
                df_rois_without_current_row = df_rois.drop(idx)
                # add the margins to the bbox
                bb1 = row[["min_x", "min_y", "max_x", "max_y"]] + margins
                # loop to all the other ROIs
                for idx2, row2 in df_rois_without_current_row.iterrows():
                    # add the margins to the bbox
                    bb2 = row2[["min_x", "min_y", "max_x", "max_y"]] + margins
                    # If intersection => merge
                    if _intersection_bbox(bb1, bb2):
                        df_rois = _fusion_bbox(df_rois, idx, idx2)
                    # test if length of df_rois >1
                    if len(df_rois) <= 1:
                        break

    return df_rois

#%%
###############################################################################
def extract_rois_core(
    sig,
    params=cfg.PARAMS_EXTRACT,
    display=False,
    verbose=False,
    **kwargs):
    
    """ Extract all Rois in the audio file
    Parameters
    ----------
    audio_path : TYPE
        DESCRIPTION. 
    params : dictionnary
        contains all the parameters to extract the rois 
    display : TYPE, optional
        DESCRIPTION. The default is False.
    verbose : TYPE, optional
        DESCRIPTION. The default is False.
    Returns
    -------
    df_rois : TYPE
        DESCRIPTION.
    """
    
    # 1. compute the spectrogram
    Sxx, tn, fn, ext = maad.sound.spectrogram(
                                    sig,
                                    params["SAMPLE_RATE"],
                                    nperseg=params["NFFT"],
                                    noverlap=params["NFFT"] // 2)

    t_resolution = tn[1] - tn[0]
    f_resolution = fn[1] - fn[0]

    if verbose:
        print("time resolution {}s".format(t_resolution))
        print("frequency resolution {}s".format(f_resolution))

    if display:

        # creating grid for subplots
        fig = plt.figure()
        fig.set_figheight(5)
        fig.set_figwidth(13)
        ax0 = plt.subplot2grid(shape=(2, 4), loc=(0, 0), colspan=1)
        ax1 = plt.subplot2grid(shape=(2, 4), loc=(1, 0), colspan=1)
        ax2 = plt.subplot2grid(shape=(2, 4), loc=(0, 1), colspan=1)
        ax3 = plt.subplot2grid(shape=(2, 4), loc=(1, 1), colspan=1)
        ax4 = plt.subplot2grid(shape=(2, 4), loc=(0, 2), rowspan=2, colspan=2)

        maad.util.plot_spectrogram(
                        Sxx,
                        extent=ext,
                        ax=ax0,
                        title="1. original spectrogram",
                        interpolation=None,
                        now=False)

    # 2. reduce the size of the spectrogram Sxx
    # Both parameters can be adapted to the audio signature that we want to select
    Sxx_reduced = resize(
                    Sxx, 
                    (Sxx.shape[0] // params["FACTOR_F"], 
                     Sxx.shape[1] // params["FACTOR_T"]),
                    anti_aliasing=True)
    
    if display:
        maad.util.plot_spectrogram(
                            Sxx_reduced,
                            log_scale=True,
                            extent=ext,
                            ax=ax1,
                            title="2. Resized spectrogram",
                            interpolation=None,
                            now=False)

    # 3. Clean spectrogram : remove background)
    Sxx_clean_reduced, _ = maad.sound.remove_background_along_axis(
                                                    Sxx_reduced,
                                                    mode=params["MODE_RMBCKG"],
                                                    N=params["N_RUNNING_MEAN"],
                                                    display=False)
    
    # 4. convert to dB
    Sxx_clean_dB_reduced = (maad.util.power2dB(Sxx_clean_reduced, db_range=96) +96)

    if display:
        maad.util.plot_spectrogram(
                        Sxx_clean_dB_reduced,
                        extent=ext,
                        log_scale=False,
                        ax=ax2,
                        title="3. cleaned spectrogram",
                        interpolation=None,
                        now=False)

    # reduce time and frequential vectors
    if params["FACTOR_T"] % 2 == 0:  # is even
        tn_reduced = (tn[np.arange(params["FACTOR_T"] // 2, 
                                   len(tn), 
                                   params["FACTOR_T"])]
                      + t_resolution/2)
    else:  # is odd
        tn_reduced = tn[np.arange(params["FACTOR_T"] // 2,
                                  len(tn), 
                                  params["FACTOR_T"])]

    if params["FACTOR_F"] % 2 == 0:  # is even
        fn_reduced = (fn[ np.arange(params["FACTOR_F"] // 2, 
                                    len(fn), 
                                    params["FACTOR_F"])]
                      + f_resolution/2)
    else:  # is odd
        fn_reduced = fn[np.arange(params["FACTOR_F"] // 2,
                                  len(fn), 
                                  params["FACTOR_F"])]

    # 7. binarization of the spectrogram to select part of the spectrogram with
    # acoustic activity
    # Both parameters can be adapted to the situation in order to take more
    # or less ROIs that are more or less large
    im_mask = maad.rois.create_mask(
                        Sxx_clean_dB_reduced,
                        mode_bin="absolute",
                        bin_h=params["MASK_PARAM1"],
                        bin_l=params["MASK_PARAM2"])

    if display:
        maad.util.plot_spectrogram(
                        im_mask,
                        extent=ext,
                        ax=ax3,
                        title="4. mask",
                        interpolation=None,
                        now=True)

    # 8. get the mask with rois (im_rois) and the bounding box for each rois (rois_bbox)
    # and an unique index for each rois => in the pandas dataframe rois
    _, df_rois = maad.rois.select_rois(im_mask, min_roi=None)
 
    # and format ROis to reduced tn and fn
    df_rois = maad.util.format_features(df_rois, tn_reduced, fn_reduced)
    # and format ROis to initial tn and fn
    df_rois = maad.util.format_features(df_rois, tn, fn)

    # Remove cut ROIs (begining and end of audio) and 1 line rois (vertical
    # and horizontal )
    # Drop ROIs with same min_x max_x and min_y max_y (=> 1 line)
    df_rois = df_rois[df_rois.min_x < df_rois.max_x]
    df_rois = df_rois[df_rois.min_y < df_rois.max_y]

    # Drop two columns
    df_rois = df_rois.drop(
        columns=["labelID", "label"])

    if verbose:
        print("\nBEFORE MERGING FOUND {} ROIS ".format(len(df_rois)))

    # Test if we found an ROI otherwise we pass to the next chunk
    if len(df_rois) > 0:

        if len(df_rois) > 1:
            # 9. Merge ROIs that are very close to each other.
            # Default, if marings = [0,0,0,0], bbox must overlap in order
            # to be merge. If margins is not null, all bbox are expanded
            # according to margins. If the expanded bbox overlapped, they
            # are merged.
            # The process is iterative which means than several ROIs can be
            # merged after several pass
            margins = [
                -round(params["MARGIN_T_LEFT"] / t_resolution),
                -round(params["MARGIN_F_BOTTOM"] / f_resolution),
                 round(params["MARGIN_T_RIGHT"] / t_resolution),
                 round(params["MARGIN_F_TOP"] / f_resolution)]
            df_rois = _merge_bbox(df_rois, margins)

        # Keep only events with duration longer than MIN_DURATION
        df_rois = df_rois[((df_rois["max_t"] - df_rois["min_t"]) > params["MIN_DURATION"])]

        if verbose:
            print("=> AFTER MERGING FOUND {} ROIS".format(len(df_rois)))

        if display:
            # Convert in dB
            X = maad.util.power2dB(Sxx, db_range=96) + 96
            kwargs = {"vmax": np.max(X)}
            kwargs.update({"vmin": np.min(X)})
            kwargs.update({"extent": ext})
            kwargs.update({"figsize": (1, 2.5)})
            maad.util.plot_spectrogram(X, ext, 
                                  log_scale=False, 
                                  ax=ax4, 
                                  title="5. Overlay ROIs")
            maad.util.overlay_rois(X, df_rois, 
                              ax=ax4, fig=fig,
                              edge_color='yellow',
                              **kwargs)
            kwargs.update({"ms": 4, "marker": "+", "fig": fig, "ax": ax4})
            # ax, fig = maad.util.overlay_centroid(X, df_rois, **kwargs)
            fig.suptitle(kwargs.pop("suptitle", ""))
            fig.tight_layout()

    return df_rois


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" 
Collection of functions to compute the features for each sound of interest also
called regions of interest (ROIs)
"""
#
# Authors:  Felix MICHAUD   <felixmichaudlnhrdt@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

#%%
# general packages
import os
from pathlib import Path

# basic packages
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

# audio package
import librosa

# Parallel processing packages
from functools import partial
from tqdm import tqdm
from concurrent import futures

# Scikit-Maad (ecoacoustics functions) package
import maad

#
from bambird import config as cfg
# cfg.get_config()
from bambird import grab_audio_to_df


#%%

###############################################################################
def compute_features(
    audio_path,
    params=cfg.PARAMS['PARAMS_FEATURES'],
    display=False,
    verbose=False):
    """ 
    Compute features such as shape (wavelets), centroid and bandwidth
    
    Parameters
    ----------
    audio_path : string
        full path the audio file
    params : dictionnary, optional
        contains all the parameters to compute the features
    display : boolean, optional
        if true, display the spectrograms and the signals at each step of 
        the process. The default is False.
    verbose : boolean, optional
        if true, print information. The default is False.
        
    Returns
    -------
    df_features : pandas dataframe
        dataframe with all the features computed for each roi found in the audio. 
    """
    
    # 1.   Load an audio
    # 2.   Bandpass the audio
    # 3.   Compute the spectrogram
    # 4.   Convert to dB
    # 5.   Remove background noise
    # 6.   Compute the features shape of the ROIs 
    # 7.   Compute the centroid and bandwidth of the ROIs
    # 8.   Add the features into a dataframe

    plt.style.use("default")

    # extract audio filename and categories from fullfilename
    path, filename_with_ext = os.path.split(audio_path)
    filename = filename_with_ext
    _, categories = os.path.split(path)
    
    if verbose:
        print("file {}".format(filename))

    try:
        # 1. load audio
        #----------------------------------------------
        sig, sr = librosa.load(audio_path, sr=params["SAMPLE_RATE"])

        # 2. bandpass filter around birds frequencies
        #----------------------------------------------
        fcut_max = min(params["HIGH_FREQ"], params["SAMPLE_RATE"] // 2 - 1)
        fcut_min = params["LOW_FREQ"]
        sig = maad.sound.select_bandwidth(sig,
                                    params["SAMPLE_RATE"],
                                    fcut=[fcut_min, fcut_max],
                                    forder=params["BUTTER_ORDER"],
                                    fname="butter",
                                    ftype="bandpass")

        # 3. compute the spectrogram
        #----------------------------------------------
        Sxx, tn, fn, ext = maad.sound.spectrogram(sig,
                                             params["SAMPLE_RATE"],
                                             nperseg=params["NFFT"],
                                             noverlap=params["NFFT"] // 2)

        # set to zero the DC value (= the first row of the spectrogram)
        Sxx[0, :] = 0
        Sxx[1, :] = 0

        if display:
            # prepare 4 figures in a row
            fig, ax = plt.subplots(2, 2, figsize=(10, 5))
            # plot the spectrogram into a bigger plot full of zeros with fixed
            # size in order to compare the spectrograms
            WIDTH_T = 20  # in s. Fix the width of the spectrogram.
            # width in pixels
            width_x = round(WIDTH_T / (tn[1] - tn[0]))
            # zero padding before and after the spectrogram if the duration time
            # is below width_x
            if int(round((width_x - Sxx.shape[1]) / 2)) > 0 :
                margin = np.zeros((Sxx.shape[0], 
                                  int(round((width_x - Sxx.shape[1]) / 2))), int)
                Sxx_to_display = np.concatenate((margin, Sxx, margin), axis=1)
            else :
                Sxx_to_display = Sxx
                
            maad.util.plot_spectrogram(Sxx_to_display,
                                  extent=[0, WIDTH_T, ext[2], ext[3]],
                                  ax=ax[0, 0],
                                  title="original ROI",
                                  now=False)

        # 4. convert to dB
        #----------------------------------------------
        Sxx_dB = maad.util.power2dB(Sxx, db_range=96) + 96

        # 5. Clean spectrogram : remove background
        #----------------------------------------------
        Sxx_clean_dB, _, _ = maad.sound.remove_background(Sxx_dB)
        Sxx_clean_dB = Sxx_clean_dB * (Sxx_clean_dB > 6)

        if display:
            maad.util.plot_spectrogram(
                np.concatenate((margin, Sxx_clean_dB, margin), axis=1),
                # [left, right, bottom, top]
                extent=[0, WIDTH_T, ext[2], ext[3]],
                log_scale=False,
                ax=ax[0, 1],
                title="cleaned ROI",
                now=False)

        # initialize a dataframe with ROIs with the coordinate of the
        # bounding box
        # !!! it does not work when using the original bbox values found during
        # the segmentation. This might be due to the different resolution which
        # creates pixel shift.

        # test if we can find 0 pixels. If not, set the min_y to 0 and max_y to
        # the vertical size of the spectrogram
        if (len(np.where(maad.sound.avg_amplitude_spectro(Sxx_clean_dB) != 0)[0])) <= 2:
            min_y = 0
            max_y = Sxx_clean_dB.shape[1]
        else:
            min_y = np.where(maad.sound.avg_amplitude_spectro(Sxx_clean_dB) != 0)[0][2]  # 2 to avoid to take indice 0 or 1
            max_y = np.where(maad.sound.avg_amplitude_spectro(Sxx_clean_dB) != 0)[0][-1]

        df_rois_for_shape = pd.DataFrame([[min_y, 0, max_y, Sxx.shape[1] - 1]], 
                                         columns=["min_y", "min_x", "max_y", "max_x"])

        # 6. Compute acoustic features (MAAD)
        #----------------------------------------------
        df_shape, params_shape = maad.features.shape_features(Sxx_clean_dB, 
                                                         resolution=params["SHAPE_RES"], 
                                                         rois=df_rois_for_shape)

        if display:
            maad.util.plot_shape(df_shape, params_shape, ax=ax[1, 0])

        # Keep columns ['min_y', 'min_x', 'max_y', 'max_x'] in case of these
        # features are not in the original dataset. If there are already
        # in the original dataset, these columns will be deleted and we keep
        # the original columns ['min_y', 'min_x', 'max_y', 'max_x'] found during
        # the extraction of the ROIs
        df_shape = maad.util.format_features(df_shape, tn, fn)

        # 7. Compute the centroid (t,f), the bandwidth and the duration of the ROI
        #-------------------------------------------------------------------------
        # initialize the dataframe df_rois with duration and centroid from the
        # size of the spectrogram
        df_rois = pd.DataFrame([[tn[-1], tn[-1] / 2]], columns=["duration_t", "centroid_t"])

        # compute the mean spectrum and smooth the spectrum with a running mean
        Sxx_clean = maad.util.dB2power(Sxx_clean_dB)
        mean_spectrum = maad.util.running_mean(maad.sound.avg_power_spectro(Sxx_clean), N=35)

        # Convert the spectrum in dB
        mean_spectrum_dB = maad.util.power2dB(mean_spectrum)

        # find the frequency position and amplitude corresponding to the
        # frequency peak
        f_peak = fn[np.argmax(mean_spectrum_dB)]
        f_peak_amplitude_dB = np.max(mean_spectrum_dB)

        # find the frequencies (- and +) at -6dB
        index = [i for i, x in enumerate(mean_spectrum_dB) if x > (f_peak_amplitude_dB-6)]
        f_peak_minus_6dB = fn[index[0]]
        f_peak_plus_6dB = fn[index[-1]]

        # add the values to the dataframe
        df_rois["peak_f"] = f_peak
        df_rois["centroid_f"] = (f_peak_plus_6dB + f_peak_minus_6dB) / 2
        df_rois["bandwidth_f"] = f_peak_plus_6dB - f_peak_minus_6dB
        df_rois["bandwidth_min_f"] = f_peak_minus_6dB
        df_rois["bandwidth_max_f"] = f_peak_plus_6dB
        
        if display:
            maad.util.plot_spectrum(mean_spectrum_dB, fn, ax=ax[1, 1])
            ax[1, 1].axvline(x=f_peak_minus_6dB, color="red", linestyle="dotted", linewidth=2)
            ax[1, 1].axvline(x=f_peak_plus_6dB, color="red", linestyle="dotted", linewidth=2)
            ax[1, 1].plot([f_peak], [f_peak_amplitude_dB], "r+")
            ax[1, 1].plot([(f_peak_plus_6dB + f_peak_minus_6dB) /2],[f_peak_amplitude_dB], "bo")
            fig.suptitle(filename + " " + categories)
            fig.tight_layout()

        # add global information about the file and categories for all rois (rows)
        df_rois.insert(0, "filename_ts", filename)
        df_rois.insert(1, "fullfilename_ts", audio_path)

        # 8. concat df_rois and df_shape
        #----------------------------------------------
        df_features = pd.concat([df_rois, df_shape], axis=1)

    except Exception as e:
        print("\n" + str(e))
        df_features = pd.DataFrame()

    return df_features

###############################################################################
def multicpu_compute_features(
                dataset, 
                params=cfg.PARAMS['PARAMS_FEATURES'],
                save_path=None,
                save_csv_filename=None,
                nb_cpu=None,
                overwrite=False,
                verbose=True): 
    """
    Parameters
    ----------
    dataset : string or pandas dataframe
        if it's a string it should be either a directory where are the ROIs
        files to process or a full path to a csv file containing a column
        "filename_ts" and a column "fullfilename_ts" with the full path to the 
        ROIS files to process
        if it's a dataframe, the dataframe should contain a column
        "filename_ts" and a column "fullfilename_ts" with the full path to the audio
        files to process. This dataframe can be obtained by called the function
        grab_audio_to_df        
    params : dictionnary, optional
        contains all the parameters to compute the features
    save_path : string, default is None
        Path to the directory where the csv file with the features will be saved    
    save_csv_filename: string, optional
        csv filename that contains all the features that will be saved. The default
        is None, meaning that the name will be automatically created from the
        parameters to compute the features as :
            'features_'
            + params["SHAPE_RES"]
            + "_NFFT"
            + str(params["NFFT"])
            + "_SR"
            + str(params["SAMPLE_RATE"])
            + ".csv"
    overwrite : boolean, optional
        if a directory already exists with the rois, if false, the process is 
        aborted, if true, new features will eventually be added in the directory and
        in the csv file.
    nb_cpu : integer, optional
        number of cpus used to compute the features. The default is None which means
        that all cpus will be used
    verbose : boolean, optional
        if true, print information. The default is False.
        
    Returns
    -------
    df_features_sorted : pandas dataframe
        dataframe containing all the features computed for each roi. 
    csv_fullfilename : string
        full path the csv file with all the features computed for each roi. 
        if the file already exists, the new features will be appended to the file.
    """

    if verbose :
        print('\n')
        print('====================== COMPUTE FEATURES ======================\n')
        
    # if dataset is a directory : 
    # > all audio files in the directory will be processed. 
    #-------------------------------------------------------------------------
    # test if dataset is a valid path with audio files
    if isinstance(dataset, pd.DataFrame) == False:
        
        # format dataset into Path
        dataset = Path(dataset)
        
        # test if it's a directory
        if os.path.isdir(dataset):
    
            # create a dataframe with all recordings in the directory
            df_rois = grab_audio_to_df(path         =dataset, 
                                       audio_format ='wav', 
                                       verbose      =verbose)
            
            # change the default name of the column fullfilename when using
            # grab_audio_to_df
            df_rois = df_rois.rename(columns={"fullfilename":"fullfilename_ts",
                                              "filename":"filename_ts"})
            
            # retrieve the origianl filename from filename_ts
            df_rois['filename'] = df_rois['filename_ts'].apply(lambda filename_ts : 
                                                                  (filename_ts.split("_", 1)[0])+'.mp3')
            
            # Default directory to save the dataframe with all features
            #----------------------------------------------------------
            if save_path is None :
                save_path = dataset
                
        # test if dataset is a valid csv file
        #------------------------------------
        elif os.path.isfile(dataset):
            # load the data from the csv
            df_rois = pd.read_csv(dataset, sep = ';')
                    
            # Default directory to save the dataframe with all features
            #----------------------------------------------------------
            if save_path is None :
                save_path = dataset.parent
            
    # if dataset is a dataframe : 
    # > read the dataframe       
    #-------------------------------------------------------------------------
    elif isinstance(dataset, pd.DataFrame): 
        df_rois = dataset.copy()
        
        # Default directory to save the dataframe with all features
        #----------------------------------------------------------
        if save_path is None:
            save_path = str(Path(df_rois['fullfilename_ts'].iloc[0]).parent.parent) 

    else:
        raise Exception(
            "WARNING: dataset must be a valid path to a directory, a csv file"
            + "or a dataframe"
        )
              
    # Default name of the csv file with features
    #-------------------------------------------------------------------------
    if save_csv_filename is None :
        save_csv_filename = (
            'features_'
            + params["SHAPE_RES"]
            + "_NFFT"
            + str(params["NFFT"])
            + "_SR"
            + str(params["SAMPLE_RATE"])
            + ".csv"
        )
                
    # format save_path into Path
    save_path = Path(save_path)
    
    try :
        # load the dataframe with all ROIs already extracted
        df_features = pd.read_csv(save_path / save_csv_filename, 
                                  sep=';')
        # create a mask to select or not the audio files that were already segmented
        mask = df_rois['filename'].isin(df_features['filename'].unique().tolist())
        
    except :
        # create an empty dataframe. It will contain all ROIs found for each
        # audio file in the directory
        df_features = pd.DataFrame()  
        
        # create a mask full of false
        mask = np.zeros(len(df_rois), bool)
    
    INITIAL_NUM_ROIS = len(df_features)

    # if necessary compute features using multicpu  
    #-----------------------------------------------
    # test if the dataframe contains files to compute the features
    if len(df_rois[~mask]) > 0 :
        
        # Test if the csv with features already exists and if we want to overwrite it.
        #-----------------------------------------------------------------------------
        file_exist = (save_path / save_csv_filename).exists()
        if (file_exist == False) or ((file_exist == True) and (overwrite == True)) : 
        
            if ((file_exist == True) and (overwrite == True))  :
                if verbose :
                    print(('{} already exists. \n'+
                          '***BE CAREFULL*** csv file will be overwritten').format(save_csv_filename)) 
        
            if verbose :
                print('Composition of the dataset : ')
                print('   -number of files : %2.0f' % len(df_rois[~mask]))
                print('   -number of categories : %2.0f' % len(df_rois[~mask].categories.unique()))
                print('   -unique categories : {}'.format(df_rois[~mask]['categories'].unique()))
            
            # Number of CPU used for the calculation. By default, set to all available
            # CPUs
            if nb_cpu is None:
                nb_cpu = os.cpu_count()
        
            #================================================================================
        
            # # define a new function with fixed parameters to give to the multicpu pool
            multicpu_func = partial(
                compute_features,
                params=params,
                display=False,
                verbose=False,
            )
        
            with tqdm(total=len(df_rois[~mask])) as pbar:
                with futures.ProcessPoolExecutor(max_workers=nb_cpu-1) as pool:
                    for df_features_temp in pool.map(
                        multicpu_func, df_rois[~mask]["fullfilename_ts"].to_list()
                    ):
                        pbar.update(1)
                        df_features = df_features.append(df_features_temp)
                                
            # Merge the result df_features with the df_rois
            #----------------------------------------------
            # Find the columns that are also in the first DataFrame
            # For instance, if columns ['min_y', 'min_x', 'max_y', 'max_x'] are in both
            # dataframe, we will keep the original value from df_data found during
            # the extraction of the ROIs
            
            df_rois.set_index(["filename_ts"], inplace=True)
            df_features.set_index(["filename_ts"], inplace=True)
            drop_col = (df_features.columns & df_rois.columns).tolist()
        
            # Merge the DataFrames based on the index and after dropping the columns
            # that are duplicated
            df_features = pd.merge(
                df_features.drop(drop_col, axis=1),
                df_rois,
                how="outer",
                left_index=True,
                right_index=True,
            )
            
            # sort filename for each categories
            #-------------------------------
            df_features_sorted = pd.DataFrame()
            for categories in df_features["categories"].unique():
                df_features_sorted = df_features_sorted.append(
                    df_features[df_features["categories"] == categories].sort_index()
                )
           
            if verbose :
                print('\n')
                print(('Features were extracted for'
                       +' {} ROIS audio files').format(len(df_features_sorted)- INITIAL_NUM_ROIS))
             
            # save features
            #-------------------------------
            if save_csv_filename is not None :
                # Set filename_ts to be the index before saving the dataframe
                try:
                    df_features_sorted.set_index(['filename_ts'], inplace=True)
                except:
                    pass  
                # save and append dataframe 
                csv_fullfilename = save_path / save_csv_filename
                df_features_sorted.to_csv(csv_fullfilename, 
                                          sep=';', 
                                          header=True)
                # reset index
                df_features_sorted.reset_index(inplace=True)
    
    # otherwise, select the features rows to return
    #----------------------------------------------
    else :        
        # reset the index
        try:
            df_features.reset_index('filename_ts', inplace = True)
        except:
            pass       
                
        # remove from df_rois the audio files that were already computed
        mask = df_features['filename'].isin(df_rois['filename'].unique().tolist())
        df_features_sorted = df_features[mask]
        csv_fullfilename = save_path / save_csv_filename
        
        if verbose :
            print("No feature needs to be computed")
            print(">>> FEATURES COMPUTING PROCESS ABORTED <<<")
            
    return df_features_sorted, csv_fullfilename



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" 
Collection of functions to query metadata from xeno-canto and download the audio
to build a dataset.
"""
#
# Authors:  Felix MICHAUD   <felixmichaudlnhrdt@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

#%%
# general packages
import os
from pathlib import Path

# basic packages
import pandas as pd
import glob

# Scikit-Maad (ecoacoustics functions) package
import maad

#
from bambird import config as cfg
# cfg.get_config()

#%%

###############################################################################
def query_xc (species_list, 
           params=cfg.PARAMS['PARAMS_XC'],
           format_time=False,
           format_date=False,
           random_seed=cfg.RANDOM_SEED, 
           verbose=False):
    """
    Query metadata from Xeno-Canto website with audiofile depending on the search terms. 
    The audio recordings metadata are grouped and stored in a dataframe.

    Parameters
    ----------
    species_list : list
        List of scientific name of birds (e.g. 'Columba palumbus')
    
    params : list
        list of search terms to perform the query
        The main seach terms are :
        - q   : quality
        - cnt : country
        - len : length
        - area : continent (europe, africa, america, asia)
        see more here : https://www.xeno-canto.org/help/search
    format_time : boolean, optional
        Time in Xeno-Canto is not always present neither correctly formated. 
        If true, time will be correctly formated to be processed as DateTime 
        format. When formating is not possible, the row is dropped. 
        The default is False
    format_date : boolean, optional
        Date in Xeno-Canto is not always present neither correctly formated. 
        If true, rows with uncorrect format of date are dropped.
    verbose : boolean, optional
        Print messages during the execution of the function. The default is False.

    Returns
    -------
    df_dataset : pandas DataFrame
        Dataframe containing all the recordings metadata matching search terms
    """
    
    # separate the genus and the species from the scientific name
    gen = []
    sp = []
    for name in species_list:
        gen.append(name.rpartition(' ')[0])
        sp.append(name.rpartition(' ')[2])

    # add the genus and the species as query parameters
    df_query = pd.DataFrame()
    df_query[0] = gen
    df_query[1] = sp

    # add the other parameters
    if params is not None :
        idx = 2
        for param in params['PARAM_XC_LIST']:
            df_query[idx] = param
            idx += 1

    # Get recordings metadata corresponding to the query
    df_dataset = maad.util.xc_multi_query(df_query,
                                     max_nb_files=params['NUM_FILES'],
                                     format_time=format_date,
                                     format_date=format_date,
                                     random_seed=random_seed,
                                     verbose=verbose)

    # test if recordings are found
    if len(df_dataset) > 0 :
        # add scientific name and categories into the dataframe
        df_dataset.insert(4, 'categories', df_dataset['gen'] + ' ' + df_dataset['sp'])        
    else:  
        if verbose :
            print("No audio recording was found with the parameters: \n {}".format(params))

    
    return df_dataset

###############################################################################
def download_xc (df_dataset,
                rootdir, 
                dataset_name, 
                csv_filename="bam_metadata.csv",
                overwrite=False,
                verbose=True):
    """
    Download the audio files from Xeno-Canto based on the input dataframe
    It will create directories for each species if needed

    Parameters
    ----------
    df_dataset : pandas DataFrame
        Dataframe containing the selected recordings metadata to be downloaded.
        It could be the output of dataset_query, or a subset of this
        dataframe
    rootdir : string
        Path to the directory where the whole dataset will be saved
    dataset_name : string
        Name of the dataset that will be created as a parent directory . 
    csv_filename : string, optional
        Name of the csv file where the dataframe (df_dataset) will be saved. if
        the file already exists, data will be appended at the end of the csv file.
        The default is bam_metadata.csv
    overwrite : boolean, optional
        Test if the directory where the audio files will be downloaded already
        exists. if True, it will download the data in the directory anyway.
        Otherwise, if False, it will not download audio files.
    verbose : boolean, optional
        Print messages during the execution of the function. The default is False.

    Returns
    -------
    df_dataset : pandas DataFrame
        Dataframe containing all the selected audio recordings that were 
        resquested as input AND downloaded. If for some reasons, some of the 
        audio recordings cannot be downloaded, they do not appear in the output
        dataframe. 
    csv_fullfilename : string
        Full path to the csv file where the dataframe (df_dataset) was saved. 
        If the file already exists, data was appended at the end of the csv file.
        /!\ the csv file contains all the audio files that were recorded NOW and
        BEFPRE while the dataframe df_dataset contains ONLY the files that
        were recorded NOW.
    """
    # format rootdir as path
    rootdir = Path(rootdir)
    
    #-------------------------------------------------------------------
    # download all the audio files into a directory with a subdirectory for each
    # species
    df_dataset = maad.util.xc_download(df           =df_dataset,
                                       rootdir      =rootdir,
                                       dataset_name =dataset_name,
                                       overwrite    =overwrite,
                                       save_csv     =True,
                                       verbose      =verbose)
    
    #-------------------------------------------------------------------
    
    df_dataset['filename']  = df_dataset['fullfilename'].apply(os.path.basename)
    df_dataset['categories']   = df_dataset['fullfilename'].apply(lambda path : 
                                                                  Path(path).parts[-2])
        
    #--------------------------------------------------------------------------    
    # test if the csv file with all the metadata already exists and append the
    # df_dataset otherwise write a new csv file
    csv_fullfilename = rootdir / dataset_name / csv_filename
    if os.path.exists(csv_fullfilename):              
       # try to read the file and add the new rows (no duplicate)
        try :
                                                                   
            # remove from df_data the audio files that were already downloaded
            mask = df_dataset['filename'].isin(pd.read_csv(os.path.join(csv_fullfilename),
                                                           sep=';',
                                                           index_col='id')['filename'].unique().tolist())
            # append the new audio to the dataframe and save it
            df_dataset[~mask].to_csv(csv_fullfilename, 
                                     sep=";", 
                                     index=True,
                                     index_label='id',
                                     header = False,
                                     mode = 'a')                                                   
        except :
            pass                                                          
    else:
        # try to create a file and add a row corresponding to the index
        try :
            df_dataset.to_csv(csv_fullfilename, 
                              sep=";", 
                              index=True, 
                              index_label = 'id') 
        except :
            pass
   
    #--------------------------------------------------------------------------
    # display information about the downloaded audio files
    if verbose :        
        if len(df_dataset)>0 :
            print('*******************************************************')
            print('number of files : %2.0f' % len(df_dataset))
            print('number of categories : %2.0f' % len(df_dataset.categories.unique()))
            print('unique categories : {}'.format(df_dataset['categories'].unique()))
            print('*******************************************************')
        else :
            print('*** WARNING *** The dataframe is empty.')  
                        
    return df_dataset, csv_fullfilename

#%%
def grab_audio_to_df (path, 
                      audio_format, 
                      verbose=False) :
    """
    
    columns_name :
        First column name corresponds to full path to the filename
        Second column name correspond to the filename alone without the extension
    """
    
    # create a dataframe with all recordings in the directory
    filelist = glob.glob(os.path.join(path,
                                      '**/*.'+audio_format), 
                         recursive=True)
    
    df_dataset = pd.DataFrame()
    for file in filelist:
        categories = Path(file).parts[-2]
        iden = Path(file).parts[-1]
            
        df_dataset = df_dataset.append({
                                      'fullfilename':file,
                                      'filename'    :Path(file).parts[-1],
                                      'categories'  :categories,
                                      'id'          :iden},
                                    ignore_index=True)
        
    # set id as index
    df_dataset.set_index('id', inplace = True)
        
    if verbose :
        if len(df_dataset)>0 :
            print('*******************************************************')
            print('number of files : %2.0f' % len(df_dataset))
            print('number of categories : %2.0f' % len(df_dataset.categories.unique()))
            print('unique categories : {}'.format(df_dataset['categories'].unique()))
            print('*******************************************************')
        else :
            print('No {} audio file was found in {}'.format(audio_format,
                                                            path))    
    
    return df_dataset

#%%
def change_path (dataset_csv,
                 old_path,
                 new_path,
                 column_name,
                 verbose = False,
                 ) :
    
    # Read the csv
    df = pd.read_csv(dataset_csv)

    try:
        df[column_name] = df[column_name].str.replace(old_path, new_path)
        done = True
    except:
        done = False
        if verbose:
            print("**WARNING*** : No {} column is present in the dataframe".format(column_name))
        return done
    
    # save the dataframe with the new paths
    df.to_csv(dataset_csv, index=False)
    
    if verbose:
        print(' DONE ')
        print("Current path is {}".format (old_path)) 
        print(">>> New path is {}".format (new_path))
        
    return done



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" 
Collection of functions to load the configuration file
"""
#
# Authors:  Felix MICHAUD   <felixmichaudlnhrdt@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

#%%
# general packages
import sys
import os

# basic packages
import yaml

import bambird

#%%

RANDOM_SEED = 1979  # Fix the random seed to be able to repeat the results

PARAMS_XC = {
    'PARAM_XC_LIST': ['len:"20-60"', 'q:">C"', 'type:"song"'],
    'NUM_FILES': 20,
    'CSV_XC_FILE': 'xc_metadata.csv'
}

PARAMS_EXTRACT = {
    "FUNC" : bambird.extract_rois_full_sig,
    # Extract Audio resampling
    "SAMPLE_RATE": 48000,  # Sampling frequency in Hz
    # Audio preprocess
    "LOW_FREQ": 250,  # Low frequency in Hz of the bandpass filter applied to the audio
    "HIGH_FREQ": 12000,  # High frequency in Hz of the bandpass filter applied to the audio
    # butterworth filter order to select the bandwidth corresponding to the ROI
    "BUTTER_ORDER": 1,
    # Max duration of the audio files that we will use to compute the features
    "AUDIO_DURATION": 30,
    # Split the audio signal of chunk with duration = SIGNAL LENGTH (in second)
    "CHUNK_DURATION": 10,
    "OVLP": 0,  # Define the overlap ratio between each chunk
    # Spectrogram
    # Mode to compute the remove_background ('mean', 'median')
    "MODE_RMBCKG": "median",
    # Number of points used to compute the running mean of the noise profil
    "N_RUNNING_MEAN": 10,
    "NFFT": 1024,  # Number of points of the spectrogram
    # Combination of parameters for ROIs extraction
    "MASK_PARAM1": 26,  # 30 37
    "MASK_PARAM2": 10,  # 20 33
    # Select and merge bbox parameters
    "MAX_RATIO_YX": 7,      # ratio Y/X between the high (Y in px) and the width (X in px) of the ROI
    "MIN_DURATION": 0.1,    # minimum event duration in s
    "MARGIN_T_LEFT": 0.2,
    "MARGIN_T_RIGHT": 0.2,
    "MARGIN_F_TOP": 250,
    "MARGIN_F_BOTTOM": 250,
    # save parameters
    "MARGIN_T": 0.1,  # time margin in s around the ROI
    "MARGIN_F": 250,  # frequency margin in Hz around the ROI
    # butterworth filter order to select the bandwidth corresponding to the ROI
    "FILTER_ORDER": 5,
}

PARAMS_FEATURES = {
    # Extract Audio resampling
    "SAMPLE_RATE": 48000,  # Sampling frequency in Hz
    # Audio preprocess
    "LOW_FREQ": 250,  # Low frequency in Hz of the bandpass filter applied to the audio
    "HIGH_FREQ": 11000,  # High frequency in Hz of the bandpass filter applied to the audio
    # butterworth filter order to select the bandwidth corresponding to the ROI
    "BUTTER_ORDER": 1,
    # Spectrogram
    "NFFT": 1024,  
    # Number of points of the spectrogram
    "SHAPE_RES": "high",
}

PARAMS_CLUSTER = {
    "FEATURES": ['shp', 'centroid_f'],  # choose the features used to cluster {'shp', 'centroid_f', 'peak_f', 'duration_t', 'bandwidth_f', 'bandwidth_min_f', 'bandwidth_max_f', 'min_f', 'max_f' } 
    "PERCENTAGE_PTS": 5,                 # minimum number of ROIs to form a cluster (in % of the total number of ROIs) {number between 0 and 1 or blank}
    "MIN_PTS": None,                     # minimum number of ROIs to form a cluster {integer or blank}
    "METHOD": "DBSCAN",                 # HDBSCAN or DBSCAN
    "SCALER": "MINMAXSCALER",           # STANDARDSCALER or ROBUSTSCALER or MINMAXSCALER
    "KEEP":   "BIGGEST",                # ALL or BIGGEST
    "EPS":    "auto"                    # set the maximum distance between elements in a single clusters {a number or 'auto'}
}

PARAMS = {
    'PARAMS_XC' : PARAMS_XC,
    'PARAMS_EXTRACT' : PARAMS_EXTRACT,
    'PARAMS_FEATURES' : PARAMS_FEATURES,
    'PARAMS_CLUSTER' : PARAMS_CLUSTER
    }

#%%

""" ===========================================================================

                    Private functions 

============================================================================"""
def _fun_call_by_name(val):
    if '.' in val:
        module_name, fun_name = val.rsplit('.', 1)
        # Restrict which modules may be loaded here to avoid safety issue
        # Put the name of the module
        assert module_name.startswith('bambird')
    else:
        module_name = '__main__'
        fun_name = val
    try:
        __import__(module_name)
    except ImportError :
        raise ("Can''t import {} while constructing a Python object".format(val))
    module = sys.modules[module_name]
    fun = getattr(module, fun_name)
    return fun

def _fun_constructor(loader, node):
    val = loader.construct_scalar(node)
    print(val)
    return _fun_call_by_name(val)

def _get_loader():
    """Add constructors to PyYAML loader."""
    loader = yaml.SafeLoader
    loader.add_constructor("!FUNC", _fun_constructor)
    return loader

""" ===========================================================================

                    Public function 

============================================================================"""

def load_config(fullfilename = None):
    """
    Load the configuration file to set all the parameters of bambird

    Parameters
    ----------
    fullfilename : string, optional
        Path to the configuration file.
        if no valid configuration file is given, the parameters are set to the
        default values.

    Returns
    -------
    PARAMS : dictionary
        Dictionary with all the parameters that are required for the bambird's
        functions
    """    
    
    global PARAMS  
    global PARAMS_XC
    global PARAMS_EXTRACT
    global PARAMS_FEATURES
    global PARAMS_CLUSTER
    
    if os.path.isfile(str(fullfilename)): 
        with open(fullfilename) as f:
            PARAMS = yaml.load(f, Loader=_get_loader())
            PARAMS_XC = PARAMS['PARAMS_XC']
            PARAMS_EXTRACT = PARAMS['PARAMS_EXTRACT']
            PARAMS_FEATURES = PARAMS['PARAMS_FEATURES']
            PARAMS_CLUSTER = PARAMS['PARAMS_CLUSTER']
    else :
        print("The config file {} could not be loaded. Default parameters are loaded".format(fullfilename))
        
    return PARAMS

def get_config() :
    PARAMS = {
        'PARAMS_XC' : PARAMS_XC,
        'PARAMS_EXTRACT' : PARAMS_EXTRACT,
        'PARAMS_FEATURES' : PARAMS_FEATURES,
        'PARAMS_CLUSTER' : PARAMS_CLUSTER
        }
    return PARAMS





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" 
Collection of functions to cluster separate the sound of interests (or regions
of interest ROIs) from the noise, compute metrics and overlay the results 
"""
#
# Authors:  Felix MICHAUD   <felixmichaudlnhrdt@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

#%%
# general packages
import os
import warnings
from pathlib import Path

# basic packages
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

# audio package
import librosa

# scikit-learn (machine learning) package
from sklearn.metrics import precision_recall_fscore_support
from sklearn.neighbors import NearestNeighbors
from sklearn.cluster import DBSCAN
from sklearn.metrics import confusion_matrix
# from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler, RobustScaler, MinMaxScaler

# Kneed package to find the knee of a curve
from kneed import KneeLocator

# HDBSCAN package (clustering)
import hdbscan

# # import metrics for Density Clustering such as DBSCAN or HDBSCAN
# spec = importlib.util.spec_from_file_location(
#     "DBCV", "../../dist-packages/DBCV/DBCV/DBCV.py")
# DBCV = importlib.util.module_from_spec(spec)
# spec.loader.exec_module(DBCV)  # the module is called DBCV

# Scikit-Maad (ecoacoustics functions) package
import maad

from bambird import config as cfg
# cfg.get_config()

warnings.filterwarnings("ignore", module="librosa")
warnings.filterwarnings("ignore", module="maad")
warnings.filterwarnings(action="ignore")

plt.rcParams.update({"figure.max_open_warning": 0})
plt.style.use("default")


# %%

###############################################################################
def _prepare_features(df_features,
                     scaler = "STANDARDSCALER",
                     features = ["shp", "centroid_f"]):
    """

    Prepare the features before clustering

    Parameters
    ----------
    df_features : pandas dataframe
        the dataframe should contain the features 
    scaler : string, optional {"STANDARDSCALER", "ROBUSTSCALER", "MINMAXSCALER"}
        Select the type of scaler uses to normalize the features.
        The default is "STANDARDSCALER".
    features : list of features, optional
        List of features will be used for the clustering. The name of the features
        should be the name of a column in the dataframe. In case of "shp", "shp"
        means that all the shpxx will be used.
        The default is ["shp","centroid_f"].

    Returns
    -------
    X : pandas dataframe
        the dataframe with the normalized features 

    """

    # select the scaler
    #----------------------------------------------
    if scaler == "STANDARDSCALER":
        scaler = StandardScaler() #
    elif scaler == "ROBUSTSCALER":
        scaler = RobustScaler()
    elif scaler == "MINMAXSCALER" :
        scaler = MinMaxScaler() 
    else :
        scaler = StandardScaler()
        print ("*** WARNING *** the scaler {} does not exist. StandarScaler was choosen".format(scaler))
    
    X = pd.DataFrame()
    f = features.copy() #copy the list otherwise the suppression of a feature is kept
    
    # Normalize the shapes
    #----------------------------------------------
    if "shp" in f :
        # with shapes
        # create a vector with X in order to scale all features together
        X = df_features.loc[:, df_features.columns.str.startswith("shp")]
        X_vect = X.to_numpy()
        X_shape = X_vect.shape
        X_vect = X_vect.reshape(X_vect.size, -1)
        X_vect = scaler.fit_transform(X_vect)
        X = X_vect.reshape(X_shape)
        # remove "shp" from the list
        f.remove('shp')
    
    X2 = pd.DataFrame()
    
    # Normalize the other features (centroid, bandwidth...)
    #-------------------------------------------------------
    # test if the features list is not null
    if len(f) > 0 :
        # add other features like frequency centroid
        X2 = df_features[f]
        # Preprocess data : data scaler
        X2 = scaler.fit_transform(X2)

    # Concatenate the features after normalization
    #-------------------------------------------------------
    if len(X) == len(X2)  :
        # create a matrix with all features after rescaling
        X = np.concatenate((X, X2), axis=1)
    elif len(X2) > len(X) :
        X2 = X
    
    return X

###############################################################################
def find_cluster(
        dataset,
        params=cfg.PARAMS['PARAMS_CLUSTER'],
        save_path=None,
        save_csv_filename=None,
        display=False,
        verbose=False):
    """

    Clustering of ROIs

    We will use DBSCAN or HDSCAN clustering method for several reasons :
        * DBSCAN does not need the number of clusters to do the clustering
        * DBSCAN is able to deal with noise and keep them outside any clusters.

    So, the goal of the clustering is to aggregate similar ROIs
    which might correspond to the main call or song of a species. If several 
    clusters are found, which means that we might have ROIs corresponding to 
    different calls and/or songs for the species, we can keep the cluster with 
    the highest number of ROIs or all the clusters.

    Parameters
    ----------
    dataset : string or pandas dataframe
        if it's a string it should be a full path to a csv file with the features
        containing a column "filename_ts" and a column "fullfilename_ts" with 
        the full path to the roi
        if it's a dataframe, the dataframe should contain the features and 
        a column "filename_ts" and a column "fullfilename_ts" with the full 
        path to the roi.    
    params : dictionnary, optional
        contains all the parameters to perform the clustering
        The default is DEFAULT_PARAMS_CLUSTER.
    save_path : string, default is None
        Path to the directory where the result of the clustering will be saved    
    save_csv_filename: string, optional
        csv filename that contains all the rois with their label and cluster number
        that will be saved. The default is cluster.csv 
    display : boolean, optional
        if true, display the features vectors, the eps and 2D representation of 
        the DBSCAN or HDBSCAN results. The default is False.
    verbose : boolean, optional
        if true, print information. The default is False.

    Returns
    -------
    df_cluster : pandas dataframe
        Dataframe with the label found for each roi.

    """

    if verbose :
        print('\n')
        print('====================== CLUSTER FEATURES ======================\n')
            
    # test if dataset not a dataframe
    if isinstance(dataset, pd.DataFrame) == False:
        # test if dataset_path is a valid csv file
        if os.path.isfile(dataset):
            # load the data from the csv
            df_features = pd.read_csv(dataset, sep=';')
            
            # # Default directory to save the dataframe with all features
            # #----------------------------------------------------------
            # if save_path is None :
            #     save_path = save_path = dataset.parent
                      
    elif isinstance(dataset, pd.DataFrame): 
        df_features = dataset.copy()
        
        # reset the index
        try:
            df_features.reset_index('filename_ts', inplace = True)
        except:
            pass     
        
        # # Default directory to save the dataframe with all features
        # #----------------------------------------------------------
        # if save_path is None:
        #     save_path = str(Path(df_features['fullfilename_ts'].iloc[0]).parent.parent)

    else:
        raise Exception(
            "WARNING: dataset must be a valid path to a csv file"
            + "or a dataframe"
        )

    # drop NaN rows
    df_features = df_features.dropna(axis=0)
    
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # HACK to DELETE in the future. For compliance with data of the article 
    # The column categories does not exit
    if ('categories' in df_features.columns) == False :
        df_features["categories"] = df_features["species"]
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    if display:
        # Prepare the plots
        fig, ax = plt.subplots(1, len(df_features.categories.unique()))
        fig.set_size_inches(len(df_features.categories.unique()) * 6, 5)

        fig2, ax2 = plt.subplots(1, len(df_features.categories.unique()))
        fig2.set_size_inches(len(df_features.categories.unique()) * 6, 5)

        fig3, ax3 = plt.subplots(1, len(df_features.categories.unique()))
        fig3.set_size_inches(len(df_features.categories.unique()) * 6, 5)
        
        # in case the number of species is 1, ax, ax2, ax3 must be converted
        # into list       
        try :
            len(ax)
        except:
            ax = [ax]

        try :
            len(ax2)
        except:
            ax2 = [ax2]
        
        try :
            len(ax3)
        except:
            ax3 = [ax3]

    # count the categories
    count = 0

    # initialize df_cluster
    #-------------------------------------------------------
    df_cluster = df_features[['filename_ts',
                              'fullfilename_ts',
                              'categories',
                              'min_f',
                              'min_t',
                              'max_f',
                              'max_t']]
    
    # add 3 others columns if they exist in df_features 
    # if the ROIs were extracted outside of this workflow, they might not exist
    # in the dataframe df_features
    if 'fullfilename' in df_features :
        df_cluster['fullfilename'] = df_features['fullfilename']
    if 'filename' in df_features :
        df_cluster['filename'] = df_features['filename']
    if 'abs_min_t' in df_features :
        df_cluster['abs_min_t'] = df_features['abs_min_t']
        
    # find the cluster for each categories separately
    #-------------------------------------------------------  
    for categories in np.sort(df_features.categories.unique()):

        # select the ROIs of the current categories
        df_single_categories = df_features[df_features["categories"] == categories]
        
        # test if the number of ROIs is higher than 2.
        # If not, it is impossible to cluster ROIs. It requires at least 3 ROIS
        if len(df_single_categories) <3 :
            df_cluster["cluster_number"] = -1 # noise
            df_cluster["auto_label"] = 0 # noise
            
            if verbose:
                print("Only {} ROIs. It requires at least 3 ROIs to perform clustering".format(
                        len(df_single_categories)))
                
        else:

            # Prepare the features of that categories
            #-------------------------------------------------------
            X = _prepare_features(df_single_categories, 
                                 scaler = params['SCALER'],
                                 features = params['FEATURES'])
    
            if display:
                # Plot the features
                ax3[count].imshow(
                    X,
                    interpolation="None",
                    cmap="viridis",
                    vmin=np.percentile(X, 10),
                    vmax=np.percentile(X, 90),
                )
                ax3[count].set_xlabel("features")
                ax3[count].set_title("Shapes")
    
            # Select the minimum of points for a cluster
            #-------------------------------------------------------
            if params["PERCENTAGE_PTS"] is not None :
                min_points = round(params["PERCENTAGE_PTS"] / 100 * len(df_single_categories))
            elif params["MIN_PTS"] is not None :
                min_points = round(params["MIN_PTS"])
            else :
                min_points = 2  
                
            if min_points < 2: min_points = 2  
    
            # automatic estimation of the maximum distance eps
            #-------------------------------------------------------
            if params["EPS"] == 'auto' :
                # Calculate the average distance between each point in the data set and
                # its N nearest neighbors (N corresponds to min_points).
                neighbors = NearestNeighbors(n_neighbors=min_points)
                neighbors_fit = neighbors.fit(X)
                distances, indices = neighbors_fit.kneighbors(X)
        
                # Sort distance values by ascending value and plot
                distances = np.sort(distances, axis=0)
                distances = distances[:, 1] 
        
                # find the knee (curvature inflexion point)
                kneedle = KneeLocator(
                    x=np.arange(0, len(distances), 1),
                    y=distances,
                    interp_method="polynomial",
                    # online = False,
                    # S=10,
                    curve="convex",
                    direction="increasing",
                )
        
                if display:
                    # plot the distance + the knee
                    ax2[count].set_xlabel("cumulative number of ROIs", fontsize=10)
                    ax2[count].set_ylabel("eps", fontsize=10)
                    ax2[count].axhline(y=kneedle.knee_y, xmin=0,xmax=len(distances), color="r")
                    ax2[count].set_title("sorted k-dist graph", fontsize=12)
                    ax2[count].plot(distances)
        
                # first find the maximum distance that corresponds to 95% of observations
                eps = kneedle.knee_y  
                
                if eps == 0:
                    eps = distances.max()
           
            # set eps manually 
            #-------------------------------------------------------
            else :
                eps = params["EPS"]
    
            # find the number of clusters and the rois that belong to the cluster
            #--------------------------------------------------------------------
            if params["METHOD"] == "DBSCAN":
                cluster = DBSCAN(eps=eps, min_samples=min_points).fit(X)
                
                if verbose:
                    print("DBSCAN eps {} min_points {} Number of soundtypes found for {} : {}".format(eps, min_points,
                            categories, np.unique(cluster.labels_).size))
                    
            elif params["METHOD"] == "HDBSCAN":
                cluster = hdbscan.HDBSCAN(
                    min_cluster_size=min_points,
                    min_samples=round(min_points / 2),
                    cluster_selection_epsilon=float(eps),
                    # cluster_selection_method = 'leaf',
                    #allow_single_cluster=True,
                ).fit(X)
                
                if verbose:
                    print("filename {}".format(df_single_categories.filename))
                    print("HDBSCAN eps {} min_points {} Number of soundtypes found for {} : {}".format(eps, min_points,
                            categories, np.unique(cluster.labels_).size))
    
                # metric DBCV
                """DBCV_score = DBCV.DBCV(X, cluster.labels_, dist_function=euclidean)
                
                print('Number of soundtypes found for {} : {} \
                      / DBCV score {:.2f} \
                      / DBCV relative {:.2f}'.format(categories,
                                                   np.unique(cluster.labels_).size,
                                                   DBCV_score,
                                                   cluster.relative_validity_))"""

            # add the label found with the clustering
            #-------------------------------------------------------
            # add the cluster label into the label's column of the dataframe
            df_cluster.loc[df_cluster["categories"] == categories, "cluster_number"] = cluster.labels_.reshape(-1, 1)
    
            # add the automatic label (SIGNAL = 1 or NOISE = 0) into the auto_label's column of
            # the dataframe
            # Test if we want to consider only the biggest or all clusters 
            # that are not noise (-1) to be signal
            if params["KEEP"] == 'BIGGEST' :
                # set by default to 0 the auto_label of all
                df_cluster.loc[df_cluster["categories"] == categories, "auto_label"] = int(0)
                # find the cluster ID of the biggest cluster that is not noise
                try :
                    biggest_cluster_ID = df_cluster.loc[(df_cluster["categories"] == categories) & (
                                                     df_cluster["cluster_number"] >= 0)]["cluster_number"].value_counts().argmax()
                    # set by default to 1 the auto_label of the biggest cluster
                    df_cluster.loc[(df_cluster["categories"] == categories) & (
                                    df_cluster["cluster_number"] == biggest_cluster_ID), "auto_label"] = int(1)
                except:
                    # if there is only noise
                    pass
                
            elif params["KEEP"] == 'ALL' :
                # set by to 0 the auto_label of the noise (cluster ID = -1)
                df_cluster.loc[(df_cluster["categories"] == categories) & (
                                df_cluster["cluster_number"] < 0), "auto_label"] = int(0)
                # set by to 1 the auto_label of the signal (cluster ID >= 0)
                df_cluster.loc[(df_cluster["categories"] == categories) & (
                                df_cluster["cluster_number"] >= 0), "auto_label"] = int(1)
            
                        
            if display:
                # display the result in 2D (2D reduction of the dimension)
                pca = PCA(n_components=2)
                principalComponents = pca.fit_transform(X)
                df_PCA = pd.DataFrame(
                    data=principalComponents,
                    columns=["principal component 1", "principal component 2"],
                )
                ax[count].set_xlabel("PC 1", fontsize=10)
                ax[count].set_ylabel("PC 2", fontsize=10)
                ax[count].set_title(categories, fontsize=12)
    
                ax[count].scatter(
                    df_PCA["principal component 1"],
                    df_PCA["principal component 2"],
                    c=cluster.labels_,
                    s=50,
                    alpha=0.8,
                )

        # increment
        count += 1
            
    # Default name of the csv file with the cluster
    #-------------------------------------------------------------------------
    if save_path is not None : 
        if save_csv_filename is None :
            save_csv_filename = ('cluster.csv')
            
            # save_csv_filename = (
            #     'cluster_'
            #     'with_'
            #     + '_'.join(str(elem) for elem in params["FEATURES"])
            #     + "_min_points_"
            #     + str(round(min_points,3))
            #     + "_eps_"
            #     + str(round(eps,3))
            #     + "_method_"
            #     + str(params["METHOD"])
            #     + "_keep_"
            #     + str(params["KEEP"])
            #     + ".csv"
            # )
                    
        # format save_path into Path
        save_path = Path(save_path)
        
        if verbose :
            print('Save csv file with cluster here {}'.format(save_path/save_csv_filename))
        
        # Set filename_ts to be the index before saving the dataframe
        try:
            df_cluster.set_index(['filename_ts'], inplace=True)
        except:
            pass  
        
        # save and append dataframe 
        csv_fullfilename = save_path / save_csv_filename
        df_cluster.to_csv(csv_fullfilename, 
                                  sep=';', 
                                  header=True)
    else:
        csv_fullfilename = None

    return df_cluster, csv_fullfilename


###############################################################################
def cluster_eval(df_cluster,
                 path_to_csv_with_gt,
                 colname_label    = 'auto_label' ,
                 colname_label_gt = 'manual_label',
                 verbose=False):

    fp_initial = []
    tp_initial = []
    precision_initial = []
    precision = []
    recall = []
    tp = []
    fp = []
    tn = []
    fn = []
    number_rois_initial = []
    number_rois_final = []
    
    df = df_cluster.copy()
    
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # HACK to DELETE in the future. For compliance with data of the article 
    # The column categories does not exit
    if ('categories' in df.columns) == False :
        df["categories"] = df["species"]
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    try : 
        # load all annotations
        df_labels = pd.read_csv(path_to_csv_with_gt, sep=';')
        try :
            df_labels.drop('species', axis=1, inplace=True)
        except:
            pass
        try : 
            df_labels.drop('code', axis=1, inplace=True)
        except:
            pass
        df_labels.set_index('filename_ts', inplace=True)
        df_labels.loc[df_labels[colname_label_gt] == '0', colname_label_gt] = 0
        df_labels.loc[df_labels[colname_label_gt] == '1', colname_label_gt] = 1
    
        # join df_label and df then drop rows with NaN
        if 'filename_ts' in df :
            df.set_index('filename_ts', inplace=True)    
        df = df.join(df_labels[colname_label_gt])
        df = df.dropna(axis=0)
        
    except :
        raise Exception("WARNING: path_to_csv_with_gt must be a valid path to a csv ")
        
    # Create a new column 'marker' with tp, tn, fp, fn
    df['marker'] = None
    #TP
    df.loc[(df[colname_label]==1) * (df[colname_label_gt]==1), 'marker'] = 'TP'
    #TN
    df.loc[(df[colname_label]==0) * (df[colname_label_gt]==0), 'marker'] = 'TN'
    #FP
    df.loc[(df[colname_label]==1) * (df[colname_label_gt]==0), 'marker'] = 'FP'
    #FN
    df.loc[(df[colname_label]==0) * (df[colname_label_gt]==1), 'marker'] = 'FN'

    # select Rois that belongs to the species depending on the clustering
    
    for categories in np.sort(df.categories.unique()):

        number_rois_initial += [len(df[df["categories"] == categories])]
        number_rois_final += [np.sum((df["categories"] == categories) & (df[colname_label] == 1))]

        fp_initial += [np.sum(df[df["categories"] == categories][colname_label_gt] == 0)]
        tp_initial += [np.sum( df[df["categories"] == categories][colname_label_gt] == 1)]

        precision_initial += [round(tp_initial[-1] / (tp_initial[-1] + fp_initial[-1]) * 100)]

        _tn, _fp, _fn, _tp = confusion_matrix(
            df.dropna()[df["categories"] == categories][colname_label_gt].to_list(),
            df.dropna()[df["categories"] ==  categories][colname_label].to_list()).ravel()
        
        tp += [_tp]
        fp += [_fp]
        tn += [_tn]
        fn += [_fn]

        if (_tp + _fp) > 0:
            precision += [round(_tp / (_tp + _fp) * 100)]
        else:
            precision += [0]
        if (_tp + _fn) > 0:
            recall += [round(_tp / (_tp + _fn) * 100)]
        else:
            recall += [0]

        if verbose:
            print(
                "Initial number of ROIs is {} / Final number of ROIs is {} => {}% reduction / noise {}% => {}%  / recall {}% ({})".format(
                    number_rois_initial[-1],
                    number_rois_final[-1],
                    round(100 - number_rois_final[-1] /
                          number_rois_initial[-1] * 100, 1),
                    100 - precision_initial[-1],
                    100 - precision[-1],
                    recall[-1],
                    categories,
                )
            )

    # dataframe with scores
    df_scores = pd.DataFrame(list(zip(np.sort(df.categories.unique()),
                                      number_rois_initial,
                                      number_rois_final,
                                      tp_initial,
                                      fp_initial,
                                      tp,
                                      fp,
                                      tn,
                                      fn,
                                      precision_initial,
                                      precision,
                                      recall)),
                             columns=['species',
                                      'number_rois_initial',
                                      'number_rois_final',
                                      'tp_initial',
                                      'fp_initial',
                                      'tp',
                                      'fp',
                                      'tn',
                                      'fn',
                                      'precision_initial',
                                      'precision',
                                      'recall'])
    # set species as index
    df_scores.set_index('species', inplace = True)
    
    if verbose:
        print("------------------------------------------------------")
        print("------->Median initial noise {:.1f}%".format(
            100-np.percentile(precision_initial, 50)))
        print("Lower outlier Initial noise  {:.1f}%".format(
            100-np.percentile(precision_initial, 95)))
        print("Higher outlier Initial noise {:.1f}%".format(
            100-np.percentile(precision_initial, 5)))
        print("------->  Median Final noise {:.1f}%".format(
            100-np.percentile(precision, 50)))
        print("Lower outlier Final noise    {:.1f}%".format(
            100-np.percentile(precision,95)))
        print("Higher outlier Final noise   {:.1f}%".format(
            100-np.percentile(precision,5)))
        print("------------------------------------------------------")
        # calculate the F1-SCORE (macro and micro)
        y_true = (df.categories * df[colname_label_gt].apply(np.int64))
        y_pred = (df.categories * df[colname_label].apply(np.int64))
        print("******************************************************")
        print("avg intial noise {:2.1f}% >>> avg final noise {:2.1f}%".format(100-np.mean(precision_initial),
                                                                    100-np.mean(precision)))
        p, r, f, _ = precision_recall_fscore_support( y_true, 
                                                      y_pred, 
                                                      average='macro')
        print("MACRO precision {:.2f} | recall {:.2f} | F {:.2f}".format(p,r,f))
        print("******************************************************")

    return df_scores, p, r, f, df.marker

###############################################################################
def cluster_eval_(df_cluster,
                 path_to_csv_with_gt,
                 colname_label    = 'auto_label' ,
                 colname_label_gt = 'manual_label',
                 verbose=False):
    """

    Evalation of the clustering (requires annotations or any other files to 
                                 compare with the result of the clustering)

    Parameters
    ----------
    df_cluster : string or pandas dataframe
        if it's a string it should be a full path to a csv file with the features
        containing a column "filename_ts" and a column "fullfilename_ts" with 
        the full path to the roi
        if it's a dataframe, the dataframe should contain the features and 
        a column "filename_ts" and a column "fullfilename_ts" with the full 
        path to the roi.    
    params : dictionnary, optional
        contains all the parameters to perform the clustering
        The default is DEFAULT_PARAMS_CLUSTER.
    display : boolean, optional
        if true, display the features vectors, the eps and 2D representation of 
        the DBSCAN or HDBSCAN results. The default is False.
    verbose : boolean, optional
        if true, print information. The default is False.

    Returns
    -------
    df_cluster : pandas dataframe
        Dataframe with the label found for each roi.

    """

    fp_initial = []
    tp_initial = []
    precision_initial = []
    precision = []
    recall = []
    tp = []
    fp = []
    tn = []
    fn = []
    number_rois_initial = []
    number_rois_final = []
    
    df = df_cluster.copy()
    
    
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # HACK to DELETE in the future. For compliance with data of the article 
    # The column categories does not exit
    if ('categories' in df.columns) == False :
        df["categories"] = df["species"]
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    try : 
        # load all annotations
        df_labels = pd.read_csv(path_to_csv_with_gt, sep=';')
        try :
            df_labels.drop('species', axis=1, inplace=True)
        except:
            pass
        try : 
            df_labels.drop('code', axis=1, inplace=True)
        except:
            pass
        df_labels.set_index('filename_ts', inplace=True)
        df_labels.loc[df_labels[colname_label_gt] == '0', colname_label_gt] = 0
        df_labels.loc[df_labels[colname_label_gt] == '1', colname_label_gt] = 1
    
        # join df_label and df then drop rows with NaN
        if 'filename_ts' in df :
            df.set_index('filename_ts', inplace=True)    
        df = df.join(df_labels[colname_label_gt])
        df = df.dropna(axis=0)
        
    except :
        raise Exception("WARNING: path_to_csv_with_gt must be a valid path to a csv ")
        
    # Create a new column 'marker' with tp, tn, fp, fn
    df['marker'] = None
    #TP
    df.loc[(df[colname_label]==1) * (df[colname_label_gt]==1), 'marker'] = 'TP'
    #TN
    df.loc[(df[colname_label]==0) * (df[colname_label_gt]==0), 'marker'] = 'TN'
    #FP
    df.loc[(df[colname_label]==1) * (df[colname_label_gt]==0), 'marker'] = 'FP'
    #FN
    df.loc[(df[colname_label]==0) * (df[colname_label_gt]==1), 'marker'] = 'FN'

    # select Rois that belongs to the categories depending on the clustering
    
    for categories in np.sort(df.categories.unique()):

        number_rois_initial += [len(df[df["categories"] == categories])]
        number_rois_final += [np.sum((df["categories"] == categories) & (df[colname_label] == 1))]

        fp_initial += [np.sum(df[df["categories"] == categories][colname_label_gt] == 0)]
        tp_initial += [np.sum( df[df["categories"] == categories][colname_label_gt] == 1)]

        precision_initial += [round(tp_initial[-1] / (tp_initial[-1] + fp_initial[-1]) * 100)]

        _tn, _fp, _fn, _tp = confusion_matrix(
            df.dropna()[df["categories"] == categories][colname_label_gt].to_list(),
            df.dropna()[df["categories"] == categories][colname_label].to_list()).ravel()
        
        tp += [_tp]
        fp += [_fp]
        tn += [_tn]
        fn += [_fn]

        if (_tp + _fp) > 0:
            precision += [round(_tp / (_tp + _fp) * 100)]
        else:
            precision += [0]
        if (_tp + _fn) > 0:
            recall += [round(_tp / (_tp + _fn) * 100)]
        else:
            recall += [0]

        if verbose:
            print(
                "Initial number of ROIs is {} / Final number of ROIs is {} => {}% reduction / noise {}% => {}%  / recall {}% ({})".format(
                    number_rois_initial[-1],
                    number_rois_final[-1],
                    round(100 - number_rois_final[-1] /
                          number_rois_initial[-1] * 100, 1),
                    100 - precision_initial[-1],
                    100 - precision[-1],
                    recall[-1],
                    categories,
                )
            )

    # dataframe with scores
    df_scores = pd.DataFrame(list(zip(np.sort(df.categories.unique()),
                                      number_rois_initial,
                                      number_rois_final,
                                      tp_initial,
                                      fp_initial,
                                      tp,
                                      fp,
                                      tn,
                                      fn,
                                      precision_initial,
                                      precision,
                                      recall)),
                             columns=['categories',
                                      'number_rois_initial',
                                      'number_rois_final',
                                      'tp_initial',
                                      'fp_initial',
                                      'tp',
                                      'fp',
                                      'tn',
                                      'fn',
                                      'precision_initial',
                                      'precision',
                                      'recall'])
    # set categories as index
    df_scores.set_index('categories', inplace = True)
    
    if verbose:
        print("------------------------------------------------------")
        print("------->Median initial noise {:.1f}%".format(
            100-np.percentile(precision_initial, 50)))
        print("Lower outlier Initial noise  {:.1f}%".format(
            100-np.percentile(precision_initial, 95)))
        print("Higher outlier Initial noise {:.1f}%".format(
            100-np.percentile(precision_initial, 5)))
        print("------->  Median Final noise {:.1f}%".format(
            100-np.percentile(precision, 50)))
        print("Lower outlier Final noise    {:.1f}%".format(
            100-np.percentile(precision,95)))
        print("Higher outlier Final noise   {:.1f}%".format(
            100-np.percentile(precision,5)))
        print("------------------------------------------------------")
        # calculate the F1-SCORE (macro and micro)
        y_true = (df.categories * df[colname_label_gt].apply(np.int64))
        y_pred = (df.categories * df[colname_label].apply(np.int64))
        print("******************************************************")
        print("avg intial noise {:2.1f}% >>> avg final noise {:2.1f}%".format(100-np.mean(precision_initial),
                                                                    100-np.mean(precision)))
        p, r, f, _ = precision_recall_fscore_support( y_true, 
                                                      y_pred, 
                                                      average='macro')
        print("MACRO precision {:.2f} | recall {:.2f} | F {:.2f}".format(p,r,f))
        print("******************************************************")

    return df_scores, p, r, f, df.marker
      

###############################################################################
def overlay_rois (cluster,
                  markers = None,
                  column_labels ='auto_label',
                  unique_labels=[0,1],
                  color_labels=['tab:red', 'tab:green', 'tab:orange', 'tab:blue', 
                                 'tab:purple','tab:pink','tab:brown','tab:olive',
                                 'tab:cyan','tab:gray','yellow'],
                  textbox_label=True,
                  params=cfg.PARAMS['PARAMS_EXTRACT'],
                  filename=None,
                  random_seed=None,
                  verbose=False,
                  **kwargs):
    
    if verbose :
        print('\n')
        print('============== OVERLAY ROIS ON THE ORIGINAL FILE ==============\n')
    
    # test if dataset not a dataframe
    if isinstance(cluster, pd.DataFrame) == False:
        
        # test if dataset_path is a valid csv file
        if os.path.isfile(cluster):
            
            # load the data from the csv
            df_cluster = pd.read_csv(cluster, index_col=0)
            df_cluster.reset_index(inplace=True)
                      
    elif isinstance(cluster, pd.DataFrame): 
        df_cluster = cluster.copy()

    else:
        raise Exception(
            "***WARNING*** cluster must be a valid path to a csv file"
            + "or a dataframe"
        )
    
    # add the markers into the dataframe df_cluster if markers is not None
    if markers is not None :
        try:
            df_cluster.set_index('filename_ts', inplace = True)
        except:
            pass
        df_cluster = df_cluster.join(markers)
        df_cluster.reset_index(inplace=True)

    # select one xeno-canto file
    #---------------------------
    if "filename" in df_cluster :
        try:
            df_cluster.set_index(['filename'],inplace=True)
        except:
            pass
    else:
        # if there is no filename column but filename_ts, extract the filename
        if "filename_ts" in df_cluster :
            df_cluster['filename'] = df_cluster['filename_ts'].apply(lambda file : file.split('_',1)[0]+'.mp3')
            df_cluster.set_index(['filename'],inplace=True)
        else :
            raise Exception(
                "***WARNING*** cluster must have a column 'filename' or 'filename_ts' "
        )
    
    # if no filename given, pick a random one
    if filename is None :
        filename = df_cluster.sample(n=1, random_state=random_seed).index

    # extract the row corresponding to the filename
    df_single_file = df_cluster.loc[filename]

    # reset the index
    df_single_file.reset_index(inplace=True)
    df_cluster.reset_index(inplace=True)

    # Xeno-Canto filename
    fullfilename = df_single_file['fullfilename'].unique()[0]
    
    # add a label column
    df_single_file['label'] = df_single_file[column_labels]
    
    if verbose :
        print('Display ROIs found in the file {}'.format(fullfilename))
        print('labels : {}'.format(df_single_file.label.values))
    
    # Load the audio and add the bounding boxes
    #------------------------------------------
    try:
        # 1. load audio
        sig, sr = librosa.load(
            fullfilename, sr=params["SAMPLE_RATE"], duration=params["AUDIO_DURATION"]
        )
        
        fig = plt.figure(figsize=kwargs.pop("figsize", 
                                         (params["AUDIO_DURATION"]/60*15,7)))
        
        ax0 = plt.subplot2grid((5, 1), (0, 0), rowspan=1)
        ax1 = plt.subplot2grid((5, 1), (1, 0), rowspan=4)

        maad.util.plot_wave(s=sig, fs=sr, ax=ax0, now = False)
        
        # 2. compute the spectrogram
        Sxx, tn, fn, ext = maad.sound.spectrogram(
            sig,
            params["SAMPLE_RATE"],
            nperseg=params["NFFT"],
            noverlap=params["NFFT"] // 2,
        )
        
       
        df_single_file['min_t'] = df_single_file['min_t']+ df_single_file['abs_min_t']
        df_single_file['max_t'] = df_single_file['max_t']+ df_single_file['abs_min_t'] 
        df_single_file = maad.util.format_features(df_single_file, tn, fn)
        
        # 3.
        # Convert in dB
        X = maad.util.power2dB(Sxx, db_range=96) + 96

        kwargs.update({"vmax": np.max(X)})
        kwargs.update({"vmin": np.min(X)})
        kwargs.update({"extent": ext})
        maad.util.plot_spectrogram(X, 
                              log_scale=False, 
                              colorbar=False,
                              ax=ax1,
                              now = False,
                              **kwargs)
        
        # 4.
        if unique_labels is None :
            unique_labels = list(df_single_file[column_labels].unique())
        
        # overlay
        maad.util.overlay_rois(im_ref=X, 
                              rois = df_single_file,
                              ax=ax1,
                              fig=fig,
                              unique_labels=unique_labels,
                              edge_color=color_labels,
                              textbox_label=textbox_label)

        fig.suptitle((df_single_file.filename.unique()
                     + " " 
                     + df_single_file.categories.unique()))
        fig.tight_layout()
        plt.show()

    except Exception as e:
        if verbose:
            print("\nERROR : " + str(e))
    
    return filename

###############################################################################
def mark_rois (markers,
               dataset_csv,
               display = False,
               verbose = False):

    if verbose :
        print('\n')
        print('============= MARK ROIS with the prefix TP TN FP FN ============\n')

        
    # test if dataset_csv is a valid csv file
    if os.path.isfile(dataset_csv):
        # load the data from the csv
        df = pd.read_csv(dataset_csv, index_col=0, sep=';')
        df.reset_index(inplace=True)           
    else:
        raise Exception(
            "***WARNING*** dataset must be a valid path to a csv file")    
   
    # test if there is already a marker column.
    # If so, stop the process
    if 'marker' in df :
        marked = False
        if verbose:
            print('***WARNING*** ROIs are already marked.'
                  +' Unmarked ROIs before marking them again')
    
    else: 
        # join markers and dataset
        try :
            df.set_index('filename_ts', inplace=True)
            df = df.join(markers)
            df.reset_index(inplace=True)
        except:
            print('***WARNING*** No filename_ts column found in the dataset \n' +
                  '>>> None of the ROIs was marked')
            return
        
        # test if there is a column fullfilename_ts in the dataset
        if 'fullfilename_ts' and 'filename_ts' in df : 
            marked = True
            for idx, row in df.iterrows():
    
                f = Path(row['fullfilename_ts'])
                filename_ts = f.parts[-1]
                path = f.parent
                
                # select the marker to add to the file
                prefix = row['marker']
                # if the marker does not exit, add 'NA'
                if pd.isna(prefix) : prefix = 'NA'
    
                # rename the file with the marker
                f_new = path / Path(prefix + '_'+ filename_ts)
                os.rename(f, f_new)  
                
                # test if the file has been renamed
                if os.path.isfile(f_new) :
                    marked = marked and True
                    # add the new filename into the dataframe df
                    df.loc[idx,'fullfilename_ts'] = str(f_new)
                    df.loc[idx,'filename_ts'] = f_new.parts[-1]
                    
                else:
                    marked = False
                    if verbose :
                        print("***WARNING*** The ROIs with the filename_ts {} could not be renamed as {}".format(f, f_new))
                    
            try :
                df.set_index('filename_ts', inplace = True)
                df.to_csv(dataset_csv, sep=';')
                df.reset_index('filename_ts')
                if verbose :
                    print(" DONE \n" +
                          ">>> Save the dataframe with marked ROIs in {}".format(dataset_csv))
            except:
                if verbose :
                    print("***WARNING*** Dataframe with marked ROIs was not saved")
        else:
            marked = False
            if verbose :
                print('***WARNING*** Either fullfilename_ts or filename_ts do not exist in the dataframe. \n' +
                      '>>> None of the ROIs was marked')
    return df, marked

###############################################################################
def unmark_rois (dataset_csv,
                 display = False,
                 verbose = False):

    if verbose :
        print('\n')
        print('========================== UNMARK ROIS ========================\n')
    
    # test if dataset_csv is a valid csv file
    if os.path.isfile(dataset_csv):
        # load the data from the csv
        df = pd.read_csv(dataset_csv, index_col=0, sep=';')
        df.reset_index(inplace=True)         
    else:
        raise Exception(
            "***WARNING*** dataset must be a valid path to a csv file"
        )    
      
    # test if there is a columns fullfilename_ts filename_ts and marker in the dataset
    if ('fullfilename_ts' and 'filename_ts' and 'marker') in df : 
        unmarked = True
        for idx, row in df.iterrows():

            f = Path(row['fullfilename_ts'])
            filename_ts = f.parts[-1]
            path = f.parent
            
            # rename the file without the marker (-3 char)
            f_new = path / filename_ts[3:]
            os.rename(f, f_new)  
            
            # test if the file has been renamed
            if os.path.isfile(f_new) :
                unmarked = unmarked and True
                # add the new filename into the dataframe df
                df.loc[idx,'fullfilename_ts'] = str(f_new)
                df.loc[idx,'filename_ts'] = f_new.parts[-1]
                
            else:
                unmarked = False
                if verbose :
                    print("***WARNING*** The ROIs with the filename_ts {} could not be renamed as {}".format(f, f_new))
        
        # drop the column marker before saving
        df.drop('marker', axis=1, inplace=True)        
        
        try :
            df.set_index('filename_ts', inplace = True)
            df.to_csv(dataset_csv, sep=';')
            df.reset_index('filename_ts')
            if verbose :
                print(" DONE \n" +
                      ">>> Save the dataframe with unmarked ROIs in {}".format(dataset_csv))
        except:
            if verbose :
                print("***WARNING*** Dataframe with unmarked ROIs was not saved")
    else:
        unmarked = False
        if verbose :
            print('***WARNING*** Either fullfilename_ts or filename_ts or marker do not exist in the dataframe. \n' +
                  '>>> None of the ROIs was unmarked')
            
    return df, unmarked

    


#!/usr/bin/env python
#
# Copyright (C) 2022 Felix Michaud <felixmichaudlnhrdt@gmail.com>
#                    Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
# License: BSD 3 clause

import os
import textwrap
from setuptools import setup, find_packages, Command
from importlib.machinery import SourceFileLoader

version = SourceFileLoader('bambird.version',
                           'bambird/version.py').load_module()

with open('README.md', 'r') as fdesc:
    long_description = fdesc.read()

class CleanCommand(Command):
    """Custom clean command to tidy up the project root.
    Deletes directories ./build, ./dist and ./*.egg-info
    From the terminal type:
        > python setup.py clean
    """
    user_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        os.system('rm -vrf ./build ./dist ./*.egg-info')

setup(
      name = 'bambird',
      version = version.__version__,  # Specified at bambird/version.py file
      packages = find_packages(),
      author = 'Felix Michaud and Sylvain Haupert',
      author_email = 'felixmichaudlnhrdt@gmail.com, sylvain.haupert@mnhn.fr',
      maintainer = 'Felix Michaud and Sylvain Haupert',
      description = 'BAM, unsupervised labelling function to extract and cluster similar animal vocalizations together',
      long_description = long_description,
      long_description_content_type='text/markdown',
      license = 'BSD 3 Clause',
      keywords = ['ecoacoustics', 'bioacoustics', 'ecology', 'dataset', 'signal processing', 'segmentation', 'features', 'clustering', 'unsupervised  labelling', 'deep learning', 'machine learning'],
      url = 'https://github.com/ear-team/bambird',
      #project_urls={'Documentation': 'https://bambird.github.io'},
      platform = 'OS Independent',
      cmdclass={'clean': CleanCommand},
      license_file = 'LICENSE',                     
      python_requires='>=3.5',
      install_requires = ['scikit-maad>=1.3.12',
                          'librosa>=0.8.0',
                          'scikit-learn>=1.0',
                          'hdbscan',
                          'matplotlib', 
                          'tqdm',
                          'kneed',
                          'pyyaml'],
      classifiers=textwrap.dedent("""
        Development Status :: 4 - Beta
        Intended Audience :: Science/Research
        License :: OSI Approved :: BSD License
        Operating System :: OS Independent
        Programming Language :: Python :: 3.5
        Programming Language :: Python :: 3.6
        Programming Language :: Python :: 3.7
        Programming Language :: Python :: 3.8
        Topic :: Scientific/Engineering :: Artificial Intelligence 
        """).strip().splitlines()
       )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" 
Collection of functions to segment the sound of interests (also called regions 
of interest (ROIs) in time and frequency.
"""
#
# Authors:  Felix MICHAUD   <felixmichaudlnhrdt@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

#%%
# general packages
import os
from pathlib import Path

# basic packages
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

# Scipy
from scipy.io import wavfile

# audio package
import librosa

# Parallel processing packages
from functools import partial
from tqdm import tqdm
from concurrent import futures

# Scikit-Maad (ecoacoustics functions) package
import maad

#
from bambird import config as cfg
from bambird import grab_audio_to_df

#%%
###############################################################################

def _save_rois(
    chunk,
    fs,
    df_rois,
    save_path,
    margins=(0.02, 50),
    filter_order=5,
    display=False):
    """ 
    Save the rois as new wave files. 
    Margins around the bbox can  be applied around the bbox in order to 
    enlarge the save bbox in time and frequency
    
    Parameters
    ----------
    chunk: numpy array
        chunk of audio which a 1d vector of scalars to be saved
    fs: integer
        Sampling frequency of the audio
    df_rois: pandas dataframe
        dataframe with all the rois to be saved. Each roi is defined by limits
        in time and frequency that is used to cut the portion of audio that
        correspond to the roi.        
    margins: list of numbers, optional
        the first numbers corresponds to the +/- margins in time (s)
        the second numbers corresponds to the +/- margins in frequency (Hz)
        the default value is (0.02s, 50Hz)
    filter_order: integer, optional
        Define the order of the bandpass filter. The default value is 5
    display: boolean, optional
        If true, display the spectrogram of the roi that is saved
    
    Returns
    ------- 
    df_rois: pandas dataframe 
        returns the dataframe with a new column "fullfilename_ts" that contains
        the full path to the rois
    """

    fullfilename_ts_list = []
    
    # format save_path into Path
    save_path = Path(save_path)
    
    # Loop over the ROIs in the dataframe
    for index, row in df_rois.iterrows():

        # get the actual full path and create an new path to store
        # the ROIs as wav file
        full_path = Path(row["fullfilename"])
        # test if the save path is not a bird directory
        # if its already a bird directory, do not nest directory
        if save_path != full_path.parent:
            #new_path = os.path.join(save_path, full_path.parts[-2])
            new_path = save_path / full_path.parts[-2]
        else:
            new_path = save_path
        # create nested directories
        if os.path.exists(new_path) == False:
            new_path.mkdir(parents=True, exist_ok=True)
            
        # A) cut the raw audio in time domain
        idx_min = int((row["min_t"] - margins[0]) * fs)
        idx_max = int((row["max_t"] + margins[0]) * fs)
        # test if index are out of the boundary
        if idx_min < 0:
            idx_min = 0
        if idx_max >= len(chunk):
            idx_max = len(chunk) - 1
        # cut in time
        chunk_rois = chunk[idx_min:idx_max]

        # B) cut the audio in the frequency domain
        fcut_min = row["min_f"] - margins[1]
        fcut_max = row["max_f"] + margins[1]
        # test if frequencies are out of the boundary
        # fcut should be >0 and <fs/2
        if fcut_min <= 0:
            fcut_min = row["min_f"]
        if fcut_max > fs / 2:
            fcut_max = row["max_f"]
        # cut in frequency
        chunk_rois = maad.sound.select_bandwidth(
            chunk_rois,
            fs,
            fcut=[fcut_min, fcut_max],
            forder=filter_order,
            fname="butter",
            ftype="bandpass",
        )

        # Let's display the spectrogram of the ROI
        if display:
            # compute the spectrogram
            X, _, _, ext = maad.sound.spectrogram(
                chunk_rois, fs, nperseg=1024, noverlap=1024 // 2
            )
            kwargs = {"vmax": np.max(X)}
            kwargs.update({"vmin": np.min(X)})
            kwargs.update({"extent": ext})
            kwargs.update({"figsize": (1, 2.5)})
            maad.util.plot_spectrogram(
                X, extent=ext, title="saved spectrogram", 
                interpolation=None, now=True
            )

        # C) Write the chunk as a wav file
        # add an index to the filename in order to save unique
        # filename => filename_ts (for time stamp)
        fullfilename_ts = str(new_path) + "/" + row["filename_ts"]
        wavfile.write(fullfilename_ts, fs, np.float32(chunk_rois))

        fullfilename_ts_list.append(fullfilename_ts)

    # # add the full path for the chunk in the dataframe
    df_rois.insert(1, "fullfilename_ts", fullfilename_ts_list)

    return df_rois



###############################################################################
def single_file_extract_rois(
    audio_path,
    fun,
    params=cfg.PARAMS['PARAMS_EXTRACT'],
    save_path=None,
    display=False,
    verbose=False):
    """ Extract all Rois in the audio file
    Parameters
    ----------
    audio_path : TYPE
        DESCRIPTION.
    fun : function
        name of the function that is called to segment the rois
    params : dictionnary
        contains all the parameters to extract the rois 
    save_path : string, default is None
        Path to the directory where the segmented rois will be saved    
    display : TYPE, optional
        DESCRIPTION. The default is False.
    verbose : TYPE, optional
        DESCRIPTION. The default is False.
    Returns
    -------
    df_rois : TYPE
        DESCRIPTION.
    """

    plt.style.use("default")

    # extract audio filename from fullfilename
    path, filename_with_ext = os.path.split(audio_path)
    filename = filename_with_ext
    # extract species and categories from fullfilename 
    try :
        _, species = os.path.split(path)
        categories = ((species.split(" ", 1)[0][0:3]).lower() + 
                (species.split(" ", 1)[1][0:3]).lower())
    except :
        _, species = os.path.split(path)
        categories = species
    
    try:
        # 1. load audio
        sig, sr = librosa.load(
            audio_path, sr=params["SAMPLE_RATE"], duration=params["AUDIO_DURATION"])

        # remove DC value
        sig -= np.mean(sig)

        # 2. bandpass filter around birds frequencies
        fcut_max = min(params["HIGH_FREQ"], params["SAMPLE_RATE"] // 2 - 1)
        fcut_min = params["LOW_FREQ"]
        sig = maad.sound.select_bandwidth(
            sig,
            params["SAMPLE_RATE"],
            fcut=[fcut_min, fcut_max],
            forder=params["BUTTER_ORDER"],
            fname="butter",
            ftype="bandpass")

        # Split signal into CHUNK_DURATION second chunks
        # Just like we did before (well, this could actually be a seperate function)
        # We also adjust the overlap (OVLP, ratio) between each chunk
        sig_splits = []
        n_points_per_chunk = int(params["CHUNK_DURATION"]* params["SAMPLE_RATE"])
        for i in range(0,len(sig), int(n_points_per_chunk * (1 - params["OVLP"]))):

            split = sig[i: i + n_points_per_chunk]

            # End of signal?
            if len(split) < n_points_per_chunk :
                break

            sig_splits.append(split)

        # print outputs
        if verbose:
            print("\n================================================================")
            print("CATEGORIES : {}".format(categories))
            print("NUMBER OF CHUNKS : {}".format(len(sig_splits)))

        # if no chunk was found, pass
        if sig_splits == 0:
            pass
        else:

            # Find ROIs and compute features of each chunk and add them into df_rois
            seconds = 0
            df_rois = pd.DataFrame()
            for chunk in sig_splits:

                # call the function fun
                current_df_rois = fun(sig=chunk, 
                                    params=params, 
                                    display=display, 
                                    verbose=verbose,
                                    kwarg={'suptitle': filename + " " + categories})
                
                # test if at least 1 ROI was found
                if len(current_df_rois) > 0 :
                    
                    # Add global information about the file and categories for 
                    # all rois (rows)
                    filename_ts = [
                        filename.split(".",1)[0] + "_" +
                        str(seconds) + "s_" + "{:02d}".format(index) + ".wav"
                        for index, _ in current_df_rois.iterrows()
                    ]
                    
                    current_df_rois.insert(0, "filename_ts", filename_ts)
                    current_df_rois.insert(1, "filename", filename)
                    current_df_rois.insert(2, "fullfilename", audio_path)
                    current_df_rois.insert(3, "species", species)
                    current_df_rois.insert(4, "categories", categories)
                    current_df_rois.insert(5, "abs_min_t", seconds)
                    current_df_rois.insert(4, "id", filename.split('.',1)[0][2:])
                    
                    # Save ROIs as raw audio file
                    if save_path is not None:                        
                        if len(current_df_rois) > 0:
                            current_df_rois = _save_rois(
                                chunk,
                                params["SAMPLE_RATE"],
                                current_df_rois,
                                save_path=save_path,
                                margins=(params["MARGIN_T"],
                                         params["MARGIN_F"]),
                                filter_order=params["FILTER_ORDER"],
                                display=False,
                            )
          
                    # Add current_df_rois to df_rois
                    if len(df_rois) > 0:
                        df_rois = df_rois.append(
                            current_df_rois, ignore_index=True)
                    else:
                        df_rois = current_df_rois

                # Keep track of the end time of each chunk
                # it there is an overlap between chunks, take it into account
                seconds += params["CHUNK_DURATION"] * (1 - params["OVLP"])

    except Exception as e:
        if verbose:
            print("\nERROR : " + str(e))
        df_rois = pd.DataFrame()

    return df_rois

###############################################################################
def multicpu_extract_rois(
    dataset, 
    params=cfg.PARAMS['PARAMS_EXTRACT'],
    save_path=None,
    save_csv_filename='rois.csv',
    overwrite=False,
    nb_cpu=None,
    verbose=True):
    """ 
    Extract all Rois in the dataset (multiple audio files)
    
    Parameters
    ----------
    dataset : string or pandas dataframe
        if it's a string it should be either a directory where are the audio
        files to process or a full path to a csv file containing a column
        "filename" and a column "fullfilename" with the full path to the audio
        files to process
        if it's a dataframe, the dataframe should contain a column
        "filename" and a column "fullfilename" with the full path to the audio
        files to process. This dataframe can be obtained by called the function
        grab_audio_to_df        
    params : dictionnary, optioanl
        contains all the parameters to extract the rois 
    save_path : string, default is None
        Path to the directory where the segmented rois will be saved    
    save_csv_filename: string, optional
        csv filename that contains all the rois that will be saved. The default
        is rois.csv
    overwrite : boolean, optional
        if a directory already exists with the rois, if false, the process is 
        aborted, if true, new rois will eventually be added in the directory and
        in the csv file.
    nb_cpu : integer, optional
        number of cpus used to segment the rois. The default is None which means
        that all cpus will be used
    verbose : boolean, optional
        if true, print information. The default is False.
        
    Returns
    -------
    df_rois_sorted : pandas dataframe
        dataframe containing of the rois found in the audio. Each roi is
        characterized by a bounding box (min_f max_f, min_t max_t)
    csv_fullfilename : string
        full path the csv file with all the rois that were segmented. if the file
        already exists, the new rois will be appended to the file.
        
    See
    ---
    grab_audio_to_df
    """
    
    if verbose :
        print('======================= EXTRACT ROIS =========================\n')

    # if Dataset is a directory : 
    # > all audio files in the directory will be processed.
    #-----------------------------------------------------
    if isinstance(dataset, pd.DataFrame) == False :
        # test if dataset is a path to a directory
        #-----------------------------------------
        if os.path.isdir(dataset):
            
            # format dataset to Path
            dataset = Path(dataset)

            # create a dataframe with all recordings in the directory
            #--------------------------------------------------------
            df_data = grab_audio_to_df (path            =dataset, 
                                        audio_format    ='mp3',
                                        verbose         =verbose)
            
            # set default save_path and save_filename
            #----------------------------------------
            if save_path is None:
                if (dataset[-1] == "/") or (dataset[-1] == "/"):
                    dataset = dataset[:-1]
                save_path = str(dataset) + "_ROIS"
            
        # test if dataset_path is a valid csv file
        #----------------------------------------
        elif os.path.isfile(dataset):
            # load the data from the csv
            df_data = pd.read_csv(dataset, sep=';')
            
            # set default save_path and save_filename
            if save_path is None:
                save_path = os.path.dirname(dataset) + "_ROIS"
                
                
    # if dataset is a dataframe : 
    # > read the dataframe       
    #--------------------------- 
    elif isinstance(dataset, pd.DataFrame): 
        df_data = dataset.copy()
        
        # set default save_path and save_filename
        if save_path is None:
            save_path = str(Path(df_data['fullfilename'].iloc[0]).parent.parent) + "_ROIS"
            
    else:
        raise Exception(
            "WARNING: dataset must be a valid path to a directory, a csv file"
            + "or a dataframe"
        )

    # Check if the output directory already exists
    #---------------------------------------------
    dir_exist = os.path.exists(save_path) 
    if (dir_exist == False) or ((dir_exist == True) and (overwrite == True)) : 
        if (dir_exist == True) and (overwrite == True):
            if verbose:
                print(("The directory {} already exists" +
                      " and will be overwritten").format(save_path))   
        
        try :
            # load the dataframe with all ROIs already extracted
            #---------------------------------------------------
            df_rois = pd.read_csv(save_path / save_csv_filename, sep=';')
            # create a mask to select or not the audio files that were already segmented
            mask = df_data['filename'].isin(df_rois['filename'].unique().tolist())
            
        except :
            # create an empty dataframe. It will contain all ROIs found for each
            # audio file in the directory
            df_rois = pd.DataFrame()  
            
            # create a mask full of false
            mask = np.zeros(len(df_data), bool)
        
        INITIAL_NUM_ROIS = len(df_rois)
                
        # Extract ROIs using multicpu  
        #-----------------------------
        # test if the dataframe contains files to segment
        if len(df_data[~mask])>0 :
            
            if verbose :
                print('Composition of the dataset : ')
                print('   -number of files : %2.0f' % len(df_data[~mask]))
                print('   -number of categories : %2.0f' % len(df_data[~mask].categories.unique()))
                print('   -unique categories : {}'.format(df_data[~mask]['categories'].unique()))
        
            # Number of CPU used for the calculation. By default, set to all available
            # CPUs
            if nb_cpu is None:
                nb_cpu = os.cpu_count()
                
            # define a new function with fixed parameters to give to the multicpu pool 
            #-------------------------------------------------------------------------        
            
            # Print the characteristics of the function used to segment the files
            if verbose :
                print(params['FUNC'])
            
            multicpu_func = partial(
                single_file_extract_rois,
                fun=params['FUNC'],
                params=params,
                save_path=save_path,
                display=False,
                verbose=False,
            )
        
            # Multicpu process
            #-------------------
            with tqdm(total=len(df_data[~mask])) as pbar:
                with futures.ProcessPoolExecutor(max_workers=nb_cpu-1) as pool:
                    for df_rois_temp in pool.map(
                        multicpu_func, df_data[~mask]["fullfilename"].to_list()
                    ):
                        pbar.update(1)
                        df_rois = df_rois.append(df_rois_temp)
            
            # sort filename for each categories
            #---------------------------------
            df_rois_sorted = pd.DataFrame()
            for categories in df_rois["categories"].unique():
                df_rois_sorted = df_rois_sorted.append(
                    df_rois[df_rois["categories"] == categories].sort_index()
                )
                        
            if verbose :
               print('\n')
               print(('{} new ROIs added in {}').format(len(df_rois_sorted)-INITIAL_NUM_ROIS,
                                                        save_path))
               
            # save rois
            #-------------------------------
            if save_csv_filename is not None :
                # Set filename_ts to be the index before saving the dataframe
                try:
                    df_rois_sorted.set_index(['filename_ts'], inplace=True)
                except:
                    pass  
                
                # test if the directory exists if not, create it 
                if os.path.exists(save_path) == False:
                    save_path.mkdir(parents=True, exist_ok=True)
                
                # save and append dataframe 
                csv_fullfilename = save_path / save_csv_filename
                df_rois_sorted.to_csv(csv_fullfilename, 
                                      sep=';', 
                                      header=True)
                # reset index
                df_rois_sorted.reset_index(inplace=True)
        
        else:
            # reset the index
            try:
                df_rois.reset_index('filename_ts', inplace = True)
            except:
                pass  
            
            # remove from df_data the audio files that were already segmented
            mask = df_rois['filename'].isin(df_data['filename'].unique().tolist())
            df_rois_sorted = df_rois[mask]
            csv_fullfilename = save_path / save_csv_filename
            
            if verbose:
                print("No audio file needs to be segmented")
                print(">>> EXTRACTION PROCESS ABORTED <<<")
            
    # The directory already exists
    #-------------------------------      
    else:
        # Read the already existed file
        csv_fullfilename = save_path / save_csv_filename
        
        try :
            df_rois_sorted = pd.read_csv(csv_fullfilename, sep=";")
        except :
            df_rois_sorted = pd.DataFrame()
            
            if verbose :
                print(('***WARNING** The csv file {0} does not exist in the directory {1}').format(
                    save_csv_filename, save_path))
                
        if verbose:
            print("The directory {} already exists".format(save_path))
            print(">>> EXTRACTION PROCESS ABORTED <<<")

    return df_rois_sorted, csv_fullfilename




# -*- coding: utf-8 -*-
""" 

"""
from .version import __version__

# -*- coding: utf-8 -*-
""" 
Labelling function
==================

The package ``bambird`` is an ensemble of functions to build a labelling function to clean audio recordings from Xeno-Canto.

Config yaml
-----------
.. autosummary::
    :toctree: generated/
    load_config

Dataset
-------
.. autosummary::
    :toctree: generated/
    query_xc
    download_xc
    grab_audio_to_df
    change_path
    
ROIs segmentation
-----------------
.. autosummary::
    :toctree: generated/
    extract_rois_core
    extract_rois_full_sig    
    single_file_extract_rois
    multicpu_extract_rois

ROIs features
-------------
.. autosummary::
    :toctree: generated/
    compute_features
    multicpu_compute_features
    
ROIs clustering
---------------
.. autosummary::
    :toctree: generated/
    find_cluster
    cluster_eval
    overlay_rois
    mark_rois
    unmark_rois

"""

from .segmentation_extract_rois_full_sig import(       
    extract_rois_full_sig,
    )

from .segmentation_extract_rois_core import(       
    extract_rois_core,
    )

from .segmentation_extract_rois_in_soundscape import(       
    extract_rois_in_soundscape,
    )

from .config import (
    load_config
    )

from .dataset import(
    grab_audio_to_df,
    change_path,
    query_xc,
    download_xc,
    )

from .segmentation import(     
    single_file_extract_rois,
    multicpu_extract_rois,
    )


from .features import(
    compute_features,
    multicpu_compute_features,
    )
                 
from .cluster import (
    find_cluster,
    cluster_eval,
    overlay_rois,
    mark_rois,
    unmark_rois
    )


__all__ = [
        # config.py
        'load_config',
        # dataset.py
        'grab_audio_to_df',
        'change_path',
        'query_xc',
        'download_xc',
        # segmentation.py                         
        'extract_rois_core',
        'extract_rois_full_sig',
        'single_file_extract_rois',
        'extract_rois_in_soundscape',
        'multicpu_extract_rois',
        # features.py
        'compute_features',
        'multicpu_compute_features',
        # cluster.py
        'find_cluster',
        'cluster_eval',
        'overlay_rois',
        'mark_rois',
        'unmark_rois'
        ]





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" 
Collection of functions to segment the sound of interests (also called regions 
of interest (ROIs) in time and frequency.
"""
#
# Authors:  Felix MICHAUD   <felixmichaudlnhrdt@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

#%%

# basic packages
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

# scikit-image package
from skimage import measure
from skimage.morphology import closing

# Scikit-Maad (ecoacoustics functions) package
import maad

# import bamx
from bambird import config as cfg

#%%
###############################################################################
def _select_rois(im_bin, min_roi=None ,max_roi=None, 
                margins=(0,0), 
                verbose=False, display=False, savefig = None, **kwargs): 
    """ 
    Select regions of interest based on its dimensions.
    
    The input is a binary mask, and the output is an image with labelled pixels. 
 
    Parameters 
    ---------- 
    im : 2d ndarray of scalars 
        Spectrogram (or image) 
         
    min_roi, max_roi : scalars, optional, default : None 
        Define the minimum and the maximum area possible for an ROI. If None,  
        the minimum ROI area is 1 pixel and the maximum ROI area is the area of  
        the image 
        
    margins : tuple, default : (0, 0)
        Before selected the ROIs, an optional closing (dilation followed by an
        erosion) is performed on the binary mask. The element used for the closing
        is defined my margins. The first number is the number of pixels along 
        y axis (frequency) while the second number is the number of pixels along 
        x axis (time). This operation will merge events that are closed to
        each other in order to create a bigger ROIs encompassing all of them
        
    verbose : boolean, optional, default is False
        print messages
         
    display : boolean, optional, default is False 
        Display the signal if True 
         
    savefig : string, optional, default is None 
        Root filename (with full path) is required to save the figures. Postfix 
        is added to the root filename. 
         
    \*\*kwargs, optional. This parameter is used by plt.plot and savefig functions 
            
        - savefilename : str, optional, default :'_spectro_after_noise_subtraction.png' 
            Postfix of the figure filename 
         
        - figsize : tuple of integers, optional, default: (4,10) 
            width, height in inches.   
         
        - title : string, optional, default : 'Spectrogram' 
            title of the figure 
         
        - xlabel : string, optional, default : 'Time [s]' 
            label of the horizontal axis 
         
        - ylabel : string, optional, default : 'Amplitude [AU]' 
            label of the vertical axis 
         
        - cmap : string or Colormap object, optional, default is 'gray' 
            See https://matplotlib.org/examples/color/colormaps_reference.html 
            in order to get all the  existing colormaps 
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',  
            'viridis'... 
         
        - vmin, vmax : scalar, optional, default: None 
            `vmin` and `vmax` are used in conjunction with norm to normalize 
            luminance data.  Note if you pass a `norm` instance, your 
            settings for `vmin` and `vmax` will be ignored. 
         
        - extent : scalars (left, right, bottom, top), optional, default: None 
            The location, in data-coordinates, of the lower-left and 
            upper-right corners. If `None`, the image is positioned such that 
            the pixel centers fall on zero-based (row, column) indices. 
         
        - dpi : integer, optional, default is 96 
            Dot per inch.  
            For printed version, choose high dpi (i.e. dpi=300) => slow 
            For screen version, choose low dpi (i.e. dpi=96) => fast 
         
        - format : string, optional, default is 'png' 
            Format to save the figure 
             
        ... and more, see matplotlib    
 
    Returns 
    ------- 
    im_rois: 2d ndarray 
        image with labels as values 
             
    rois: pandas DataFrame 
        Regions of interest with future descriptors will be computed. 
        Array have column names: ``labelID``, ``label``, ``min_y``, ``min_x``,
        ``max_y``, ``max_x``,
        Use the function ``maad.util.format_features`` before using 
        centroid_features to format of the ``rois`` DataFrame 
        correctly.
        
    Examples 
    -------- 
    
    Load audio recording compute the spectrogram in dB.
    
    >>> s, fs = maad.sound.load('../data/cold_forest_daylight.wav')
    >>> Sxx,tn,fn,ext = maad.sound.spectrogram (s, fs, fcrop=(0,20000), display=True)           
    >>> Sxx_dB = maad.util.power2dB(Sxx) +96
    
    Smooth the spectrogram
    
    >>> Sxx_dB_blurred = maad.sound.smooth(Sxx_dB)
    
     Using image binarization, detect isolated region in the time-frequency domain with high density of energy, i.e. regions of interest (ROIs).
    
    >>> im_bin = maad.rois.create_mask(Sxx_dB_blurred, bin_std=1.5, bin_per=0.5, mode='relative')
    
    Select ROIs from the binary mask.
    
    >>> im_rois, df_rois = maad.rois.select_rois(im_bin, display=True)
    
    We detected the background noise as a ROI, and that multiple ROIs are mixed in a single region. To have better results, it is adviced to preprocess the spectrogram to remove the background noise before creating the mask.
    
    >>> Sxx_noNoise = maad.sound.median_equalizer(Sxx)
    >>> Sxx_noNoise_dB = maad.util.power2dB(Sxx_noNoise)     
    >>> Sxx_noNoise_dB_blurred = maad.sound.smooth(Sxx_noNoise_dB)        
    >>> im_bin2 = maad.rois.create_mask(Sxx_noNoise_dB_blurred, bin_std=6, bin_per=0.5, mode='relative') 
    >>> im_rois2, df_rois2 = maad.rois.select_rois(im_bin2, display=True)
    
    """ 
 
    # test if max_roi and min_roi are defined 
    if max_roi is None:  
        # the maximum ROI is set to the aera of the image 
        max_roi=im_bin.shape[0]*im_bin.shape[1] 
         
    if min_roi is None: 
        # the min ROI area is set to 1 pixel 
        min_roi = 1 
    
    if verbose :
        print(72 * '_') 
        print('Automatic ROIs selection in progress...') 
        print ('**********************************************************') 
        print ('  Min ROI area %d pix² | Max ROI area %d pix²' % (min_roi, max_roi)) 
        print ('**********************************************************') 
        
    # merge ROIS
    if sum(margins) !=0 :
        footprint = np.ones((margins[0]*2+1,margins[1]*2+1))
        im_bin = closing(im_bin, footprint)            
 
    labels = measure.label(im_bin)    #Find connected components in binary image 
    rprops = measure.regionprops(labels) 
     
    rois_bbox = [] 
    rois_label = [] 
     
    for roi in rprops: 
         
        # select the rois  depending on their size 
        if (roi.area >= min_roi) & (roi.area <= max_roi): 
            # get the label 
            rois_label.append(roi.label) 
            # get rectangle coordonates            
            rois_bbox.append (roi.bbox)     
                 
    im_rois = np.isin(labels, rois_label)    # test if the indice is in the matrix of indices 
    im_rois = im_rois* labels 
     
    # create a list with labelID and labelName (None in this case) 
    rois_label = list(zip(rois_label,['unknown']*len(rois_label))) 
    
    # test if there is a roi
    if len(rois_label)>0 :
        # create a dataframe rois containing the coordonates and the label 
        rois = np.concatenate((np.asarray(rois_label), np.asarray(rois_bbox)), axis=1) 
        rois = pd.DataFrame(rois, columns = ['labelID', 'label', 'min_y','min_x','max_y', 'max_x']) 
        # force type to integer 
        rois = rois.astype({'label': str,'min_y':int,'min_x':int,'max_y':int, 'max_x':int}) 
        # compensate half-open interval of bbox from skimage 
        rois.max_y -= 1 
        rois.max_x -= 1 
        
    else :
        rois = []    
        rois = pd.DataFrame(rois, columns = ['labelID', 'label', 'min_y','min_x','max_y', 'max_x']) 
        rois = rois.astype({'label': str,'min_y':int,'min_x':int,'max_y':int, 'max_x':int}) 
     
    # Display 
    if display :  
        ylabel =kwargs.pop('ylabel','Frequency [Hz]') 
        xlabel =kwargs.pop('xlabel','Time [sec]')  
        title  =kwargs.pop('title','Selected ROIs')  
        extent=kwargs.pop('extent',None)
                
        if extent is None : 
            xlabel = 'pseudotime [points]'
            ylabel = 'pseudofrequency [points]'
         
        # randcmap = rand_cmap(len(rois_label)) 
        # cmap   =kwargs.pop('cmap',randcmap)  
        cmap   =kwargs.pop('cmap','tab20') 
         
        _, fig = maad.util.plot2d (
                                im_rois, 
                                extent = extent,
                                title  = title,  
                                ylabel = ylabel, 
                                xlabel = xlabel,
                                cmap   = cmap, 
                                **kwargs) 
        # SAVE FIGURE 
        if savefig is not None :  
            dpi   =kwargs.pop('dpi',96) 
            format=kwargs.pop('format','png')  
            filename=kwargs.pop('filename','_spectro_selectrois')                 
            filename = savefig+filename+'.'+format 
            fig.savefig(filename, bbox_inches='tight', dpi=dpi, format=format, 
                        **kwargs)  
            
    return im_rois, rois 

#%%
###############################################################################
def extract_rois_in_soundscape(
    sig,
    params=cfg.PARAMS_EXTRACT,
    display=False,
    verbose=False,
    **kwargs):
    """ Extract all Rois in the audio file
    Parameters
    ----------
    sig : TYPE
        DESCRIPTION.
    params : dictionnary
        contains all the parameters to extract the rois 
    display : TYPE, optional
        DESCRIPTION. The default is False.
    verbose : TYPE, optional
        DESCRIPTION. The default is False.
    Returns
    -------
    df_rois : TYPE
        DESCRIPTION.
    """

    # 1. compute the spectrogram
    Sxx, tn, fn, ext = maad.sound.spectrogram(
                                sig,
                                params["SAMPLE_RATE"],
                                nperseg=params["NFFT"],
                                noverlap=params["NFFT"] // 2,
                                flims=[params["LOW_FREQ"], params["HIGH_FREQ"]])

    t_resolution = tn[1] - tn[0]
    f_resolution = fn[1] - fn[0]

    if verbose:
        print("\ntime resolution {}s".format(t_resolution))
        print("frequency resolution {}s".format(f_resolution))

    if display:
        # creating grid for subplots
        fig = plt.figure()
        fig.set_figheight(5)
        fig.set_figwidth(13)
        ax0 = plt.subplot2grid(shape=(2, 4), loc=(0, 0), colspan=1)
        ax1 = plt.subplot2grid(shape=(2, 4), loc=(1, 0), colspan=1)
        ax2 = plt.subplot2grid(shape=(2, 4), loc=(0, 1), colspan=1)
        ax3 = plt.subplot2grid(shape=(2, 4), loc=(1, 1), colspan=1)
        ax4 = plt.subplot2grid(shape=(2, 4), loc=(0, 2), rowspan=2, colspan=2)

        maad.util.plot_wave(sig, fs=params["SAMPLE_RATE"], ax=ax0)

        maad.util.plot_spectrogram(
            Sxx,
            extent=ext,
            ax=ax1,
            title="1. original spectrogram",
            interpolation=None,
            now=False)
        
    # 3. convert to dB
    Sxx_dB = maad.util.power2dB(Sxx, db_range=96) + 96
    
    # 2. Clean spectrogram : remove background)
    Sxx_clean_dB, _ = maad.sound.remove_background_along_axis(Sxx_dB,
                                                      mode=params["MODE_RMBCKG"],
                                                      N=params["N_RUNNING_MEAN"],
                                                      display=False)

    if display:
        maad.util.plot_spectrogram(
            Sxx_clean_dB,
            extent=ext,
            log_scale=False,
            ax=ax2,
            title="2. cleaned spectrogram",
            interpolation='none',
            now=False,
            vmin = 0,
            vmax = np.percentile(Sxx_clean_dB,99.9)
        )
    
    # 4. snr estimation to threshold the spectrogram
    _,bgn,snr,_,_,_ = maad.sound.spectral_snr(maad.util.dB2power(Sxx_clean_dB))
    if verbose :
        print('BGN {}dB / SNR {}dB'.format(bgn,snr))
        
    # 5. binarization of the spectrogram to select part of the spectrogram with
    # acoustic activity
    # Both parameters can be adapted to the situation in order to take more
    # or less ROIs that are more or less large        
    
    im_mask = maad.rois.create_mask(
        Sxx_clean_dB,
        mode_bin="absolute",
        bin_h=snr+params["MASK_PARAM1"],
        bin_l=snr+params["MASK_PARAM2"]
    )

    if display:
        maad.util.plot_spectrogram(
            im_mask,
            extent=ext,
            ax=ax3,
            title="3. mask",
            interpolation=None,
            now=True,
        )

    # 6. get the mask with rois (im_rois) and the bounding box for each rois (rois_bbox)
    # and an unique index for each rois => in the pandas dataframe rois
    _, df_rois = _select_rois(im_mask, min_roi=None)
    if verbose:
        print("BEFORE MERGING FOUND {} ROIS ".format(len(df_rois)))
        
    margins = (round(params["MARGIN_F_BOTTOM"] / f_resolution),
               round(params["MARGIN_T_LEFT"] / t_resolution)) 
    _, df_rois = _select_rois(im_mask, min_roi=None, margins = margins)
    
    # and format ROis to initial tn and fn
    df_rois = maad.util.format_features(df_rois, tn, fn)
    
    # 7. Remove cut ROIs (begining and end of audio) and 1 line rois (vertical
    # and horizontal )
    df_rois = df_rois[df_rois.min_x < df_rois.max_x]
    df_rois = df_rois[df_rois.min_y < df_rois.max_y]

    # Drop two columns
    df_rois = df_rois.drop(columns=["labelID", "label"])

    # Test if we found an ROI otherwise we pass to the next chunk
    if len(df_rois) > 0:

        # if len(df_rois) > 1:
        #     # 8. Merge ROIs that are very close to each other.
        #     # Default, if marings = [0,0,0,0], bbox must overlap in order
        #     # to be merge. If margins is not null, all bbox are expanded
        #     # according to margins. If the expanded bbox overlapped, they
        #     # are merged.
        #     # The process is iterative which means than several ROIs can be
        #     # merged after several pass
        #     margins = [
        #         -round(params["MARGIN_T_LEFT"] / t_resolution),
        #         -round(params["MARGIN_F_BOTTOM"] / f_resolution),
        #          round(params["MARGIN_T_RIGHT"] / t_resolution),
        #          round(params["MARGIN_F_TOP"] / f_resolution),
        #     ]

        #     num_rois = len(df_rois) + 1
        #     while len(df_rois) < num_rois:
        #         num_rois = len(df_rois)
        #         df_rois = _merge_bbox(df_rois, margins)

        # Keep only events with duration longer than MIN_DURATION
        df_rois = df_rois[((df_rois["max_t"]-df_rois["min_t"])>params["MIN_DURATION"])]
        
        # remove rois with ratio >max_ratio_xy (they are mostly artefact 
        # such as wind, ain or clipping)
        # add ratio x/y
        df_rois['ratio_yx'] = (df_rois.max_y -df_rois.min_y) / (df_rois.max_x -df_rois.min_x) 
        if params["MAX_RATIO_YX"] is not None :
            df_rois = df_rois[df_rois['ratio_yx'] < params["MAX_RATIO_YX"]] 

        if verbose:
            print("=> AFTER MERGING FOUND {} ROIS".format(len(df_rois)))
        
        if display:
            # Convert in dB
            X = maad.util.power2dB(Sxx, db_range=96) + 96
            kwargs.update({"vmax": np.max(X)})
            kwargs.update({"vmin": np.min(X)})
            kwargs.update({"extent": ext})
            kwargs.update({"figsize": (1, 2.5)})
            maad.util.plot_spectrogram(
                X, ext, log_scale=False, ax=ax4, title="5. Overlay ROIs"
            )
            maad.util.overlay_rois(X, df_rois,
                              edge_color='yellow',
                              ax=ax4, fig=fig, **kwargs)
            kwargs.update(
                {"ms": 4, "marker": "+", "fig": fig, "ax": ax4})
            # ax, fig = maad.util.overlay_centroid(X, df_rois, **kwargs)
            fig.suptitle(kwargs.pop("suptitle", ""))
            fig.tight_layout()
    
    return df_rois